"""This module defines the `jsonclassify` function."""
from __future__ import annotations
from typing import Any, Callable, Optional, Union, cast
from inspect import signature, getmro
from .jobject import JObject
from .ctx import Ctx, CtxCfg
from .fdef import FDef, FStore, FType
from .types import Types
from .modifiers.instanceof_modifier import InstanceOfModifier
from .jfield import JField
from .isjsonclass import isjsonobject
from .ograph import OGraph
from .odict import OwnedDict
from .olist import OwnedList
from .outils import (
    to_owned_dict, to_owned_list, unowned_copy_dict, unowned_copy_list
)
from .keypath import (
    concat_keypath, initial_keypath, single_key_args, compound_key_args,
    keypath_split
)
from .excs import (AbstractJSONClassException, ValidationException,
                         JSONClassResetError, JSONClassResetNotEnabledError,
                         UnlinkableJSONClassException,
                         UnauthorizedActionException)
from .orm import ORMObject



def __init__(self: JObject, **kwargs: dict[str, Any]) -> None:
    """Initialize a new jsonclass object from keyed arguments or a dict.
    This method is suitable for accepting web and malformed inputs. Eager
    validation and transformation are applied during the initialization
    process.
    """
    if self.__class__.cdef.jconf.abstract:
        raise AbstractJSONClassException(self.__class__)
    self._set_initial_status()
    for field in self.__class__.cdef.fields:
        if field.fdef.fstore != FStore.CALCULATED:
            setattr(self, field.name, None)
        if field.fdef.fstore == FStore.LOCAL_KEY:
            transformer = self.__class__.cdef.jconf.ref_name_strategy
            local_key = transformer(field)
            if field.fdef.ftype == FType.LIST:
                setattr(self, local_key, to_owned_list(self, [], local_key))
            else:
                setattr(self, local_key, None)
            self._local_keys.add(local_key)
            self._local_key_map[local_key] = field.name
    self._set(single_key_args(kwargs), fill_blanks=True)
    self._keypath_set(compound_key_args(kwargs))
    try:
        self._graph.put(self)
    except UnlinkableJSONClassException:
        pass


def jsonobject_set(self: JObject, **kwargs: dict[str, Any]) -> JObject:
    """Set object values in a batch. This method is suitable for web and
    fraud inputs. This method takes accessor marks into consideration,
    means readonly and internal field values will be just ignored.
    Writeonce fields are accepted only if the current value is None. This
    method triggers eager validation and transform. This method returns
    self, thus you can chain calling with other instance methods.
    """
    self._set(single_key_args(kwargs), fill_blanks=False)
    self._keypath_set(compound_key_args(kwargs))
    return self


def _set(self: JObject,
         kwargs: dict[str, Any], fill_blanks: bool = False) -> None:
    """Set values of a jsonclass object internally."""
    ctxcfg = CtxCfg(fill_dest_blanks=fill_blanks, all_fields=False)
    ctx = Ctx.rootctx(self, ctxcfg, kwargs)
    InstanceOfModifier(self.__class__).transform(ctx)


def _keypath_set(self: JObject, kwargs: dict[str, Any]) -> None:
    for key, value in kwargs.items():
        items = keypath_split(key)
        dest = getattr(self, items[0])
        fdef = self.__class__.cdef.field_named(items[0]).types.fdef
        used_items = [items[0]]
        self._set_to_container(dest, items[1:], value, fdef, used_items)


def _set_to_container(self: JObject,
                      dest: Any,
                      items: list[str],
                      value: Any,
                      fdef: FDef,
                      used_items: list[str]) -> None:
    if fdef.ftype == FType.INSTANCE:
        if dest is None:
            raise ValueError(f"value in {'.'.join(used_items)} is None")
        if len(items) == 1:
            dest._set({items[0]: value}, fill_blanks=False)
        else:
            dest._keypath_set({'.'.join(items): value})
    elif fdef.ftype == FType.LIST:
        if dest is None:
            raise ValueError(f"value in {'.'.join(used_items)} is None")
        if len(items) == 1:
            dest[int(items[0])] = value
        else:
            fdef = fdef.item_types.fdef
            self._set_to_container(dest[int(items[0])], items[1:], value, fdef, used_items + [items[0]])
    elif fdef.ftype == FType.DICT:
        if dest is None:
            raise ValueError(f"value in {'.'.join(used_items)} is None")
        if len(items) == 1:
            dest[items[0]] = value
        else:
            fdef = fdef.item_types.fdef
            self._set_to_container(dest[items[0]], items[1:], value, fdef, used_items + [items[0]])


def update(self: JObject, **kwargs: dict[str, Any]) -> JObject:
    """Update object values in a batch. This method is suitable for
    internal inputs. This method ignores accessor marks, thus you can
    update readonly and internal values through this method. Writeonce
    doesn't have effect on this method. You can change writeonce fields'
    value freely in this method. This method does not trigger eager
    validation and transform. You should pass valid and final form values
    through this method. This method returns self, thus you can chain
    calling with other instance methods.
    """
    unallowed_keys = (set(kwargs.keys())
                      - set(self.__class__.cdef.update_names))
    unallowed_keys_length = len(unallowed_keys)
    if unallowed_keys_length > 0:
        keys_list = "', '".join(list(unallowed_keys))
        raise ValueError(f"'{keys_list}' not allowed in "
                         f"{self.__class__.__name__}.")
    for key, item in kwargs.items():
        setattr(self, key, item)
    return self


def tojson(self: JObject,
           ignore_writeonly: bool = False,
           reverse_relationship: bool = False,
           output_null: bool | None = None) -> dict[str, Any]:
    """Convert this JSON Class object to JSON dict.

    Args:
        ignore_writeonly (bool): Whether ignore writeonly marks on
        fields. Be careful when setting it to True.

        reverse_relationship (bool): Whether including reverse relationship in
        json outputs.

        output_null (bool): Whether output null instead of leaving unexisting
        fields for null values.

    Returns:
        dict[str, Any]: A dict represents this object's JSON object.
    """
    self._can_read_check()
    ctxcfg = CtxCfg(ignore_writeonly=ignore_writeonly,
                    reverse_relationship=reverse_relationship,
                    output_null=output_null)
    ctx = Ctx.rootctx(self, ctxcfg)
    return InstanceOfModifier(self.__class__).tojson(ctx)


def pick(self: JObject, picks: list[str]) -> None:
    """
    Partial picks some items that from JSON Class to return, it will work when
    convert this JSON Class object to JSON dict.
    """
    self.is_partial = True
    self._partial_picks = picks


def omit(self: JObject, omits: list[str]) -> None:
    """
    Partial omits some items that from JSON Class to return, it will work when
    convert this JSON Class object to JSON dict.
    """
    self.is_partial = True
    if hasattr(self, '_partial_picks'):
        picks = self._partial_picks
    else:
        picks = self.cdef._field_names
    for o in omits:
        picks.remove(o)
    self._partial_picks = picks

def validate(self: JObject, all_fields: Optional[bool] = None) -> JObject:
    """Validate the jsonclass object's validity. Raises ValidationException
    on validation failed.

    Args:
        validate_all_fields (bool): Whether continue validation to fetch more
        error messages after the first error is found. This is useful when you
        are building a frontend form and want to display detailed messages.

    Returns:
        None: upon successful validation, returns nothing.
    """
    ctxcfg = CtxCfg(all_fields=all_fields)
    ctx = Ctx.rootctx(self, ctxcfg)
    InstanceOfModifier(self.__class__).validate(ctx)
    return self


@property
def is_valid(self: JObject) -> bool:
    """Test whether the jsonclass object is valid or not. This method
    triggers object validation.

    Returns:
        bool: the validity of the object.
    """
    try:
        self.validate(all_fields=False)
    except ValidationException:
        return False
    return True


def opby(self: JObject, operator: Any) -> JObject:
    """Assigns an operator object to the current object. This is used for
    operator based validation.

    Args:
        operator (Any): Anything that is operator defined by user.

    Returns:
        JObject: The JObject itself is returned.
    """
    setattr(self, '_operator', operator)
    if operator is None:
        return self
    if self.is_new:
        class_def = self.__class__.cdef
        for field in class_def.assign_operator_fields:
            fidname = self.__class__.cdef.jconf.ref_name_strategy(field)
            if field.fdef.operator_assign_transformer is not None:
                transformer = field.fdef.operator_assign_transformer
                params_len = len(signature(transformer).parameters)
                if params_len == 1:
                    setattr(self, fidname, transformer(operator)._id)
                elif params_len == 2:
                    setattr(self, fidname, transformer(operator, self)._id)
            else:
                setattr(self, fidname, operator._id)
    return self


@property
def is_new(self: JObject) -> bool:
    """This property is true if this object is newly created and not persisted
    yet.
    """
    return self._is_new


@property
def is_modified(self: JObject) -> bool:
    """This property indicates this object is modified and it's a new version
    comparing to the persisted one in the database. Calling save will cause
    modified fields to be written into the persistance storage.
    """
    return self._is_modified


@property
def is_partial(self: JObject) -> bool:
    """This property indicates this object is not a full version of the item it
    represents in the database. Only part of fields are fetched from the
    database.
    """
    return self._is_partial


@property
def is_deleted(self: JObject) -> bool:
    """This property records whether this object is deleted."""
    return self._is_deleted


@property
def modified_fields(self: JObject) -> tuple[str]:
    """A tuple of string represents the modified field names which need
    validation.
    """
    return tuple(self._modified_fields)


@property
def persisted_modified_fields(self: JObject) -> tuple[str]:
    """This is similar to `modified_fields`. This property doesn't include
    temporary fields. Thus, use `persisted_modified_fields` when updating the
    database.
    """
    retval: list[str] = []
    cdef = self.__class__.cdef
    for name in self._modified_fields:
        if cdef.field_named(name).fdef.fstore != FStore.TEMP:
            retval.append(name)
    return tuple(retval)


@property
def previous_values(self: JObject) -> dict[str, Any]:
    """This field records values to be reset to. This is only used for fields
    with compare mark or `reset_all_fields` is defined in the class
    configuration.
    """
    return self._previous_values


@property
def unlinked_objects(self: JObject) -> dict[str, list[JObject]]:
    """Unlinked objects of this jsonclass object.
    """
    return self._unlinked_objects


def reset(self: JObject) -> None:
    """Reset this object to it's unmodified status.
    """
    if not self.__class__.cdef.jconf.reset_all_fields:
        raise JSONClassResetNotEnabledError()
    if self.is_new:
        raise JSONClassResetError()
    for k, v in self.previous_values.items():
        setattr(self, k, v)
        self._modified_fields = set()
        self._is_modified = False
        self._previous_values = {}


def save(self: JObject,
         validate_all_fields: bool = False,
         skip_validation: bool = False) -> JObject:
    """Save this object into database. This will not write if no storage
    modifier is used.
    """
    self._can_create_or_update_check()
    if self.is_new:
        self._run_on_create_callbacks()
    else:
        self._run_on_update_callbacks()
    if not skip_validation:
        self.validate(all_fields=validate_all_fields)
    self._set_on_save()
    self._database_write()
    for _, lst in self.unlinked_objects.items():
        for item in lst:
            lst.remove(item)
            item.save(validate_all_fields=validate_all_fields,
                      skip_validation=skip_validation)
    return self


def delete(self: JObject) -> JObject:
    """Delete this object from database and clear linked relationships with
    delete rule.
    """
    self._can_delete_check()
    self._run_on_delete_callbacks()
    self._orm_delete()
    return self


def restore(self: JObject) -> JObject:
    """Restore this object from database and setup lost relationships with
    delete rule.
    """
    self._orm_restore()
    return self


def complete(self: JObject) -> JObject:
    """Fetch missing field values from the underlying ORM.
    """
    if self.is_partial:
        saved_modified_fields = list(self._modified_fields)
        saved_is_modified = self._is_modified
        self._orm_complete()
        self._is_modified = saved_is_modified
        self._modified_fields = saved_modified_fields
    return self


def include(self: JObject, field_name: str) -> JObject:
    """Fetch objects on reference field.
    """
    field = self.__class__.cdef.field_named(field_name)
    if field.fdef.ftype == FType.INSTANCE:
        cls = cast(ORMObject, field.fdef.inst_cls)
    elif field.fdef.ftype == FType.LIST:
        cls = field.fdef.item_types.fdef.inst_cls
    modified_fields = self.modified_fields
    is_modified = self.is_modified
    if field.fdef.fstore == FStore.LOCAL_KEY:
        if field.fdef.ftype == FType.INSTANCE:
            rkes = self.__class__.cdef.jconf.ref_name_strategy
            ridname = rkes(field)
            rid = getattr(self, ridname)
            result = cls.id(rid).exec()
            setattr(self, field.name, result)
        elif field.fdef.ftype == FType.LIST:
            rkes = self.__class__.cdef.jconf.ref_name_strategy
            ridname = rkes(field)
            rids = getattr(self, ridname)
            result = cls.ids(rids).exec()
            setattr(self, field.name, result)
    elif field.fdef.fstore == FStore.FOREIGN_KEY:
        if field.fdef.use_join_table:
            ffield = field.foreign_field
            rkes = cls.cdef.jconf.ref_name_strategy
            idref = rkes(ffield)
            result = cls.find(**{idref: [self._id]}).exec()
            setattr(self, field.name, result)
        elif field.fdef.ftype == FType.INSTANCE:
            ffield = field.foreign_field
            rkes = cls.cdef.jconf.ref_name_strategy
            idref = rkes(ffield)
            result = cls.one(**{idref: self._id}).exec()
            setattr(self, field.name, result)
        elif field.fdef.ftype == FType.LIST:
            ffield = field.foreign_field
            rkes = cls.cdef.jconf.ref_name_strategy
            idref = rkes(ffield)
            result = cls.find(**{idref: [self._id]}).exec()
            setattr(self, field.name, result)
    setattr(self, '_modified_fields', set(modified_fields))
    setattr(self, '_is_modified', is_modified)
    return self

def _orm_complete(self: JObject) -> JObject:
    """ORM method override. Fetch missing field values and assign to this object.
    """
    pass


@property
def _data_dict(self: JObject) -> dict[str, Any]:
    """A dict which is a subview of __dict__ that only contains public data
    field items.
    """
    retval = {}
    for k, v in self.__dict__.items():
        if not k.startswith('_'):
            if k not in self._local_keys:
                retval[k] = v
    return retval


def _mark_new(self: JObject) -> None:
    """Mark the jsonclass object as a new object."""
    setattr(self, '_is_new', True)
    setattr(self, '_is_modified', False)
    setattr(self, '_modified_fields', set())
    setattr(self, '_previous_values', {})


def _mark_unmodified(self: JObject) -> None:
    """Mark this jsonclass object as an unmodified object."""
    setattr(self, '_is_new', False)
    setattr(self, '_is_modified', False)
    setattr(self, '_modified_fields', set())
    setattr(self, '_previous_values', {})


def _set_initial_status(self: JObject) -> None:
    """Set the initial status of the JSON class object."""
    self._mark_new()
    setattr(self, '_is_partial', False)
    setattr(self, '_is_deleted', False)
    setattr(self, '_previous_values', {})
    setattr(self, '_local_keys', set())
    setattr(self, '_local_key_map', {})
    setattr(self, '_unlinked_objects', {})
    setattr(self, '_link_keys', {})
    setattr(self, '_unlink_keys', {})
    setattr(self, '_graph', OGraph())
    setattr(self, '_operator', None)


def _mark_not_new(self: JObject) -> None:
    """Mark the jsonclass object as not a new object."""
    setattr(self, '_is_new', False)


def _link_local_keys(self: JObject, fname: str, key: str | int) -> None:
    field = self.__class__.cdef.field_named(fname)
    ids_name = self.__class__.cdef.jconf.ref_name_strategy(field)
    if getattr(self, ids_name) is None:
        setattr(self, ids_name, [])
    getattr(self, ids_name).append(key)


def _unlink_local_keys(self: JObject, fname: str, key: str | int) -> None:
    field = self.__class__.cdef.field_named(fname)
    ids_name = self.__class__.cdef.jconf.ref_name_strategy(field)
    if getattr(self, ids_name) is None:
        return
    getattr(self, ids_name).remove(key)


def _add_link_key(self: JObject, fname: str, key: str | int) -> None:
    if self._link_keys.get(fname) is None:
        self._link_keys[fname] = []
    self._link_keys[fname].append(key)


def _add_unlink_key(self: JObject, fname: str, key: str | int) -> None:
    if self._unlink_keys.get(fname) is None:
        self._unlink_keys[fname] = []
    self._unlink_keys[fname].append(key)


def _add_unlinked_object(self: JObject,
                         field_name: str,
                         obj: JObject) -> None:
    """Add an object into unlinked objects pool."""
    if not self._unlinked_objects.get(field_name):
        self._unlinked_objects[field_name] = []
    if obj not in self._unlinked_objects[field_name]:
        self._unlinked_objects[field_name].append(obj)


def _del_unlinked_object(self: JObject,
                         field_name: str,
                         obj: JObject) -> None:
    """Remove an object from unlinked objects pool."""
    if not self._unlinked_objects.get(field_name):
        self._unlinked_objects[field_name] = []
    if obj in self._unlinked_objects[field_name]:
        self._unlinked_objects[field_name].remove(obj)


def _clear_unlinked_object(self: JObject) -> None:
    """Clear and reset all unlinked objects."""
    self._unlinked_objects = {}


def _set_on_save(self: JObject) -> None:
    """Update fields with setonsave marks if this object is modified. This
    is a graph operation. Objects chained with the saving object will also
    get setonsave called and saved.
    """
    modifier = InstanceOfModifier(self.__class__)
    ctx = Ctx.rootctx(self, CtxCfg())
    modifier.serialize(ctx)


def _clear_temp_fields(self: JObject) -> None:
    for field in self.__class__.cdef.fields:
        if field.fdef.fstore == FStore.TEMP:
            setattr(self, field.name, None)


def _database_write(self: JObject) -> None:
    pass


def _orm_delete(self: JObject) -> None:
    pass


def _orm_restore(self: JObject) -> None:
    pass


def _can_cu_check_common(self: JObject,
                         callbacks: list[Types | Callable],
                         action: str) -> None:
    if len(callbacks) == 0:
        return
    from .types import Types
    operator = getattr(self, '_operator')
    if operator is None:
        raise UnauthorizedActionException('no operator')
    for callback in callbacks:
        if isinstance(callback, Types):
            ctx = Ctx.rootctx(self, CtxCfg())
            tval = callback.modifier.transform(ctx)
            try:
                callback.modifier.validate(ctx.nval(tval))
            except ValidationException:
                raise UnauthorizedActionException(f'cannot {action}')
        else:
            result = callback(self, operator)
            if result is not None and result is not True:
                if isinstance(result, str):
                    raise UnauthorizedActionException(result)
                else:
                    raise UnauthorizedActionException(f'cannot {action}')


def _can_create_or_update_check(self: JObject) -> None:
    if self.is_new:
        self._can_cu_check_common(
            self.__class__.cdef.jconf.can_create, 'create')
    else:
        self._can_cu_check_common(
            self.__class__.cdef.jconf.can_update, 'update')


def _can_delete_check(self: JObject) -> None:
    self._can_cu_check_common(
        self.__class__.cdef.jconf.can_delete, 'delete')


def _can_read_check(self: JObject) -> None:
    self._can_cu_check_common(
        self.__class__.cdef.jconf.can_read, 'read')


def _run_on_create_callbacks(self: JObject) -> None:
    for callback in self.__class__.cdef.jconf.on_create:
        params_len = len(signature(callback).parameters)
        if params_len == 1:
            callback(self)
        else:
            callback(self, getattr(self, '_operator'))


def _run_on_update_callbacks(self: JObject) -> None:
    for callback in self.__class__.cdef.jconf.on_update:
        params_len = len(signature(callback).parameters)
        if params_len == 1:
            callback(self)
        else:
            callback(self, getattr(self, '_operator'))


def _run_on_delete_callbacks(self: JObject) -> None:
    for callback in self.__class__.cdef.jconf.on_delete:
        params_len = len(signature(callback).parameters)
        if params_len == 1:
            callback(self)
        else:
            callback(self, getattr(self, '_operator'))


@property
def _id(self: JObject) -> str | int | None:
    field = self.__class__.cdef.primary_field
    if not field:
        return None
    return getattr(self, field.name)


@property
def _previous_id(self: JObject) -> str | int | None:
    if self.is_new:
        return self._id
    field = self.__class__.cdef.primary_field
    if not field:
        return None
    if self.is_modified:
        if field.name in self.modified_fields:
            return self.previous_values[field.name]
    return self._id


def __is_private_attr__(name: str) -> bool:
    """Returns true if the attribute name indicates private attribute."""
    return name.startswith('_')


def __setattr__(self: JObject, name: str, value: Any) -> None:
    # use original setattr for private fields
    if __is_private_attr__(name):
        self.__original_setattr__(name, value)
        return
    # use special method for local keys
    if name in self._local_keys:
        if value == getattr(self, name):
            return
        field_name = self._local_key_map[name]
        field = self.__class__.cdef.field_named(field_name)
        if field.fdef.ftype == FType.INSTANCE:
            if value is None:
                self.__original_setattr__(name, value)
                setattr(self, field_name, None)
            else:
                # temporarily set to none if key is modified
                # in the future, may query object from graph
                self.__original_setattr__(name, value)
                setattr(self, field_name, None)
        elif field.fdef.ftype == FType.LIST:
            olist = to_owned_list(self, value or [], name)
            self.__original_setattr__(name, olist)
            if (value is None) or (value is []):
                setattr(self, field_name, [])
            else:
                new_list = []
                curvals = getattr(self, field_name)
                if curvals is None:
                    curvals = []
                for item in value:
                    existitem = next((v for v in curvals if v._id == item), None)
                    if existitem is not None:
                        new_list.append(existitem)
                setattr(self, field_name, new_list)
        if not self.is_new:
            setattr(self, '_is_modified', True)
            self._modified_fields.add(field_name)
    # use original setattr for non JSON class fields
    try:
        field = self.__class__.cdef.field_named(name)
    except ValueError:
        self.__original_setattr__(name, value)
        return
    # this is a JSON class field attribute
    if field.fdef.fstore == FStore.CALCULATED:
        if field.fdef.setter is None:
            raise Exception('do not set to readonly calculation field')
        else:
            if callable(field.fdef.setter):
                field.fdef.setter(value, self)
            else:
                ctx = Ctx.rootctxp(self, name, None, value)
                field.fdef.setter.modifier.transform(ctx)
            return
    if hasattr(self, name) and value == getattr(self, name):
        return
    # track modified and previous value
    if not self.is_new:
        setattr(self, '_is_modified', True)
        self._modified_fields.add(name)
        if self.__class__.cdef.jconf.reset_all_fields or \
                field.fdef.has_reset_modifier:
            if name not in self.previous_values:
                self.previous_values[name] = getattr(self, name)
    # make list and dict assignments owned and monitored
    if isinstance(value, list):
        value = to_owned_list(self, value, name)
    if isinstance(value, dict):
        value = to_owned_dict(self, value, name)
    if field.fdef.is_ref:
        if hasattr(self, name):
            self.__unlink_field__(field, getattr(self, name))
        self.__original_setattr__(name, value)
        if field.fdef.fstore == FStore.LOCAL_KEY:
            rkes = self.__class__.cdef.jconf.ref_name_strategy
            rname = rkes(field)
            if field.fdef.ftype == FType.INSTANCE:
                if value is None:
                    self.__original_setattr__(rname, None)
                if isjsonobject(value):
                    self.__original_setattr__(rname, value._id)
            elif field.fdef.ftype == FType.LIST:
                if value is None:
                    olist = to_owned_list(self, [], rname)
                    self.__original_setattr__(rname, olist)
                else:
                    keys = []
                    for item in value:
                        if item is not None:
                            keys.append(item._id)
                    olist = to_owned_list(self, keys, rname)
                    self.__original_setattr__(rname, olist)
        self.__link_field__(field, value)
    else:
        self.__original_setattr__(name, value)


def __getattribute__(self: JObject, name: str, default: Any = None) -> Any:
    """This getattr takes calculated fields into account.
    """
    if name.startswith('_'):
        cls = getmro(type(self))[-2]
        return super(cls, self).__getattribute__(name)
    cdef = self.__class__.cdef
    if name in cdef.calc_field_names:
        getter = cdef.field_named(name).fdef.getter
        if callable(getter):
            return getter(self)
        else:
            ctx = Ctx.rootctx(self, CtxCfg(), self)
            return getter.modifier.transform(ctx)
    cls = getmro(type(self))[-2]
    return super(cls, self).__getattribute__(name)


def __odict_will_change__(self: JObject, odict: OwnedDict) -> None:
    # record previous value
    name = initial_keypath(odict.keypath)
    field = self.__class__.cdef.field_named(name)
    if self.__class__.cdef.jconf.reset_all_fields or \
            field.fdef.has_reset_modifier:
        if field.fdef.has_linked:
            return
        if name not in self.previous_values:
            if field.fdef.ftype == FType.DICT:
                self.previous_values[name] = unowned_copy_dict(
                    getattr(self, name))
            if field.fdef.ftype == FType.LIST:
                self.previous_values[name] = unowned_copy_list(
                    getattr(self, name))


def __odict_add__(self, odict: OwnedDict, key: str, val: Any) -> None:
    if isinstance(val, dict):
        odict[key] = to_owned_dict(self, val,
                                   concat_keypath(odict.keypath, key))
    if isinstance(val, list):
        odict[key] = to_owned_list(self, val,
                                   concat_keypath(odict.keypath, key))
    # record modified
    if not self.is_new:
        setattr(self, '_is_modified', True)
        self._modified_fields.add(odict.keypath)


def __odict_del__(self, odict: OwnedDict, val: Any) -> None:
    # record modified
    if not self.is_new:
        setattr(self, '_is_modified', True)
        self._modified_fields.add(odict.keypath)


def __olist_will_change__(self, olist: OwnedList) -> None:
    # record previous value
    name = initial_keypath(olist.keypath)
    try:
        field = self.__class__.cdef.field_named(name)
    except:
        field = None
    if not field:
        return
    if self.__class__.cdef.jconf.reset_all_fields or \
            field.fdef.has_reset_modifier:
        if field.fdef.has_linked:
            return
        if name not in self.previous_values:
            if field.fdef.ftype == FType.DICT:
                self.previous_values[name] = unowned_copy_dict(
                    getattr(self, name))
            if field.fdef.ftype == FType.LIST:
                self.previous_values[name] = unowned_copy_list(
                    getattr(self, name))


def __olist_add__(self: JObject, olist: OwnedList, idx: int, val: Any) -> None:
    cdef = self.__class__.cdef
    if olist.keypath in self._local_keys:
        fname = self._local_key_map[olist.keypath]
        if not self._is_new and fname not in self._modified_fields:
            self._modified_fields.add(fname)
            self._is_modified = True
        return
    try:
        field = cdef.field_named(olist.keypath)
    except ValueError:
        field = None
    if field is not None and field.fdef.is_ref:
        self.__link_field__(field, [val])
        if field.fdef.fstore == FStore.LOCAL_KEY:
            rkes = cdef.jconf.ref_name_strategy
            rk = rkes(field)
            rlist = getattr(self, rk)
            if len(rlist) != len(olist):
                if val._id is None:
                    raise ValueError('a referenced object must have a valid primary key')
                rlist.insert(idx, val._id)
    if isinstance(val, dict):
        olist[idx] = to_owned_dict(self, val,
                                   concat_keypath(olist.keypath, idx))
    if isinstance(val, list):
        olist[idx] = to_owned_list(self, val,
                                   concat_keypath(olist.keypath, idx))
    # record modified
    if not self.is_new:
        setattr(self, '_is_modified', True)
        self._modified_fields.add(olist.keypath)


def __olist_del__(self: JObject, olist: OwnedList, val: Any) -> None:
    cdef = self.__class__.cdef
    if olist.keypath in self._local_keys:
        fname = self._local_key_map[olist.keypath]
        if not self._is_new and fname not in self._modified_fields:
            self._modified_fields.add(fname)
            self._is_modified = True
        flist = getattr(self, fname)
        if isinstance(flist, list):
            if len(flist) != len(olist):
                # TODO: reorder flist here
                # TODO: replace the underneath implementation
                match = next((v for v in flist if v._id == val), "PLACEHOLDER")
                if match != "PLACEHOLDER":
                    flist.remove(match)
        return
    try:
        field = cdef.field_named(olist.keypath)
    except ValueError:
        field = None
    if field and field.fdef.is_ref:
        self.__unlink_field__(field, [val])
        if field.fdef.fstore == FStore.LOCAL_KEY:
            ## TODO: replace the underneath implementation
            rkes = cdef.jconf.ref_name_strategy
            rk = rkes(field)
            rlist = getattr(self, rk)
            if len(rlist) != len(olist):
                if val is not None and val._id in rlist:
                    rlist.remove(val._id)
    # record modified
    if not self.is_new:
        setattr(self, '_is_modified', True)
        self._modified_fields.add(olist.keypath)


def __olist_sor__(self, olist: OwnedList) -> None:
    # TODO: sort local keys here
    # record modified
    if not self.is_new:
        setattr(self, '_is_modified', True)
        self._modified_fields.add(olist.keypath)


def __unlink_field__(self: JObject, field: JField, value: Any) -> None:
    items: list[JObject] = []
    if field.fdef.ftype == FType.INSTANCE:
        if not isjsonobject(value):
            return
        items = [value]
    if field.fdef.ftype == FType.LIST:
        if not isinstance(value, list):
            return
        items = list(value)
    for item in items:
        self._add_unlinked_object(field.name, item)
        other_field = field.foreign_field
        if other_field is None:
            return
        if other_field.fdef.ftype == FType.INSTANCE:
            if getattr(item, other_field.name) is self:
                item.__original_setattr__(other_field.name, None)
                of = other_field
                if of.fdef.fstore == FStore.LOCAL_KEY:
                    tsfm = item.__class__.cdef.jconf.ref_name_strategy
                    item.__original_setattr__(tsfm(other_field), None)
                    item._modified_fields.add(other_field.name)
                item._add_unlinked_object(other_field.name, self)
        elif other_field.fdef.ftype == FType.LIST:
            other_list = getattr(item, other_field.name)
            if isinstance(other_list, list):
                if self in other_list:
                    other_list.remove(self)
                    item._add_unlinked_object(other_field.name, self)


def __link_field__(self: JObject, field: JField, value: Any) -> None:
    items: list[JObject] = []
    if field.fdef.ftype == FType.INSTANCE:
        if not isjsonobject(value):
            return
        items = [value]
    if field.fdef.ftype == FType.LIST:
        if not isinstance(value, list):
            return
        items = value
    for item in items:
        self._del_unlinked_object(field.name, item)
        other_field = field.foreign_field
        if other_field is None:
            return
        if other_field.fdef.ftype == FType.INSTANCE:
            if getattr(item, other_field.name) != self:
                setattr(item, other_field.name, self)
                item._del_unlinked_object(other_field.name, self)
        elif other_field.fdef.ftype == FType.LIST:
            if not isinstance(getattr(item, other_field.name), list):
                setattr(item, other_field.name, [self])
                item._del_unlinked_object(other_field.name, self)
            else:
                if self not in getattr(item, other_field.name):
                    getattr(item, other_field.name).append(self)
                    item._del_unlinked_object(other_field.name, self)
        self.__link_graph__(item)


def __link_graph__(self: JObject, other: JObject) -> None:
    """
    """
    try:
        if not self._graph.has(self):
            self._graph.put(self)
    except UnlinkableJSONClassException:
        pass
    self._graph = self._graph.merged_graph(other._graph)


def jsonclassify(class_: type) -> type[JObject]:
    """Make a declared class into JSON class.

    Args:
        class_ (type): A class that user declared.

    Returns:
        JObject: A class that confirms to `JObject`.
    """
    # do not install methods for subclasses
    if hasattr(class_, '__is_jsonclass__'):
        return class_
    # type marks
    class_.__is_jsonclass__ = True
    # public methods
    class_.__init__ = __init__
    class_.set = jsonobject_set
    class_.update = update
    class_.tojson = tojson
    class_.pick = pick
    class_.omit = omit
    class_.validate = validate
    class_.is_valid = is_valid
    class_.opby = opby
    class_.is_new = is_new
    class_.is_modified = is_modified
    class_.is_partial = is_partial
    class_.is_deleted = is_deleted
    class_.modified_fields = modified_fields
    class_.persisted_modified_fields = persisted_modified_fields
    class_.previous_values = previous_values
    class_.unlinked_objects = unlinked_objects
    class_.reset = reset
    class_.save = save
    class_.delete = delete
    class_.restore = restore
    class_.complete = complete
    class_.include = include
    # protected methods
    class_._set = _set
    class_._keypath_set = _keypath_set
    class_._set_to_container = _set_to_container
    class_._orm_complete = _orm_complete
    class_._data_dict = _data_dict
    class_._mark_new = _mark_new
    class_._mark_unmodified = _mark_unmodified
    class_._set_initial_status = _set_initial_status
    class_._mark_not_new = _mark_not_new
    class_._link_local_keys = _link_local_keys
    class_._unlink_local_keys = _unlink_local_keys
    class_._add_link_key = _add_link_key
    class_._add_unlink_key = _add_unlink_key
    class_._add_unlinked_object = _add_unlinked_object
    class_._del_unlinked_object = _del_unlinked_object
    class_._clear_unlinked_object = _clear_unlinked_object
    class_._set_on_save = _set_on_save
    class_._clear_temp_fields = _clear_temp_fields
    class_._database_write = _database_write
    class_._orm_delete = _orm_delete
    class_._orm_restore = _orm_restore
    class_._can_cu_check_common = _can_cu_check_common
    class_._can_create_or_update_check = _can_create_or_update_check
    class_._can_delete_check = _can_delete_check
    class_._can_read_check = _can_read_check
    class_._run_on_create_callbacks = _run_on_create_callbacks
    class_._run_on_update_callbacks = _run_on_update_callbacks
    class_._run_on_delete_callbacks = _run_on_delete_callbacks
    class_._id = _id
    class_._previous_id = _previous_id
    # private methods
    class_.__original_setattr__ = class_.__setattr__
    class_.__setattr__ = __setattr__
    class_.__getattribute__ = __getattribute__
    class_.__odict_will_change__ = __odict_will_change__
    class_.__odict_add__ = __odict_add__
    class_.__odict_del__ = __odict_del__
    class_.__olist_will_change__ = __olist_will_change__
    class_.__olist_add__ = __olist_add__
    class_.__olist_del__ = __olist_del__
    class_.__olist_sor__ = __olist_sor__
    class_.__unlink_field__ = __unlink_field__
    class_.__link_field__ = __link_field__
    class_.__link_graph__ = __link_graph__
    return class_



from __future__ import annotations
from typing import Any, Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleAt:
    a: Optional[Any] = types.any.at("name")
    ca: Optional[Any] = types.any.at(lambda: "name")
    ta: Optional[Any] = types.any.at(types.default("name"))



"""module for alnum modifier."""
from __future__ import annotations
from ..fdef import FDef, FType
from .modifier import Modifier


class AnyModifier(Modifier):
    """Field marked with any modifier can be any value."""

    def define(self, fdef: FDef) -> None:
        fdef._ftype = FType.ANY



"""module for len modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class LenModifier(Modifier):
    """Len modifier return the length of value."""

    def transform(self, ctx: Ctx) -> int:
        is_str_or_list = isinstance(ctx.val, list) or type(ctx.val) is str
        return len(ctx.val) if is_str_or_list else ctx.val



"""module for inputvalue modifier."""
from ..fdef import FDef, EnumInput
from .modifier import Modifier


class InputValueModifier(Modifier):
    """Input value modifier tweaks enum modifier's behavior."""

    def define(self, fdef: FDef) -> None:
        if fdef.enum_input is None:
            fdef._enum_input = EnumInput.VALUE
        else:
            fdef._enum_input = fdef.enum_input | EnumInput.VALUE
        return



from __future__ import annotations
from unittest import TestCase
from tests.classes.user_onupdate import (UserOnupdate, UserOnupdateD,
                                         UserOnupdateT, UserOnupdateZ,
                                         check_value, check_args)


class TestOnUpdate(TestCase):

    def test_onupdate_is_not_triggered_for_new_objects(self):
        val = check_value()
        user = UserOnupdate(name='A', age=20)
        user.age = 5
        user._set_on_save()
        self.assertEqual(check_value(), val)

    def test_onupdate_is_not_triggered_for_unmodified_fields(self):
        val = check_value()
        user = UserOnupdate(name='A', age=20)
        setattr(user, "_is_new", False)
        user.name = 'B'
        user._set_on_save()
        self.assertEqual(check_value(), val)

    def test_onupdate_is_triggered_for_modified_fields(self):
        val = check_value()
        user = UserOnupdate(name='A', age=20)
        setattr(user, "_is_new", False)
        user.age = 25
        user._set_on_save()
        self.assertEqual(check_value(), val + 1)
        self.assertEqual(check_args(), (-1, 25))

    def test_onupdate_accepts_two_args_old_and_new(self):
        val = check_value()
        user = UserOnupdateD(name='A', age=20)
        setattr(user, "_is_new", False)
        user.age = 25
        user._set_on_save()
        self.assertEqual(check_value(), val + 1)
        self.assertEqual(check_args(), (20, 25))

    def test_onupdate_accepts_three_args_old_new_and_context(self):
        val = check_value()
        user = UserOnupdateT(name='A', age=35)
        setattr(user, "_is_new", False)
        user.age = 45
        user._set_on_save()
        self.assertEqual(check_value(), val + 1)
        self.assertEqual(check_args(), (80, 90))

    def test_onupdate_accepts_zero_args(self):
        val = check_value()
        user = UserOnupdateZ(name='A', age=20)
        setattr(user, "_is_new", False)
        user.age = 25
        user._set_on_save()
        self.assertEqual(check_value(), val + 1)
        self.assertEqual(check_args(), (-50, -50))



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.any_article import AnyArticle


class TestAny(TestCase):

    def test_any_generates_from_type_hint(self):
        article = AnyArticle(title=True, content=5)
        self.assertEqual(article._data_dict, {'title': True, 'content': 5})

    def test_any_does_not_raise_if_value_is_any_type(self):
        article = AnyArticle(title=True, content=5)
        article.validate()

    def test_any_raises_if_any_type_is_required(self):
        article = AnyArticle(title='True', content=None)
        self.assertRaisesRegex(ValidationException, 'value required', article.validate)



from __future__ import annotations
from datetime import datetime
from unittest import TestCase

from tests.classes.super_datetime import SuperDateTime


class TestTobohour(TestCase):

    def test_tobohour_transforms_datetime_into_the_time_of_beginning_of_hour(self):
        d = SuperDateTime(dtbh=datetime(2021, 10, 11, 17, 37, 27))
        self.assertEqual(d.dtbh, datetime(2021,10, 11, 17, 0))

    def test_tobohour_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(stbh="12345")
        self.assertEqual(s.stbh, "12345")



"""module for truncate modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class TruncateModifier(Modifier):
    """Truncate modifier truncates value."""

    def __init__(self, mlen: int | Callable | Types) -> None:
        self.mlen = mlen

    def transform(self, ctx: Ctx) -> Any:
        return ctx.val[:self.resolve_param(self.mlen, ctx)] if hasattr(ctx.val, '__len__') else ctx.val



"""module for cascade modifier."""
from ..fdef import FDef, DeleteRule
from .modifier import Modifier


class CascadeModifier(Modifier):
    """Cascade modifier marks a relationship's delete rule as cascade."""

    def define(self, fdef: FDef) -> None:
        fdef._delete_rule = DeleteRule.CASCADE



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_list import SimpleList
from tests.classes.typed_list import TypedList
from tests.classes.list_quiz import ListQuiz
from tests.classes.nullable_list import NullableList
from tests.classes.list_dict import ListDict


class TestListOf(TestCase):

    def test_listof_raises_if_value_is_not_list(self):
        slst = SimpleList(list=5)
        self.assertRaises(ValidationException, slst.validate)

    def test_listof_raises_if_one_of_values_does_not_match_inner(self):
        slst = SimpleList(list=[5, 4, 3, 2, 1.5])
        self.assertRaises(ValidationException, slst.validate)

    def test_listof_does_not_raise_if_all_values_match_inner(self):
        slst = SimpleList(list=[5, 4, 3, 2, 1])
        slst.validate()

    def test_listof_accepts_raw_type(self):
        slst = SimpleList(list=[5, 4, 3, 2, 1.5])
        self.assertRaises(ValidationException, slst.validate)
        slst1 = SimpleList(list=[5, 4, 3, 2, 1])
        slst1.validate()

    def test_listof_accepts_types_type(self):
        slst = TypedList(list=[5, 4, 3, 2, 1.5])
        self.assertRaises(ValidationException, slst.validate)
        slst1 = TypedList(list=[5, 4, 3, 2, 1])
        slst1.validate()

    def test_listof_does_not_allow_none_for_raw_typed_list(self):
        record1 = SimpleList(list=[5, 4, None])
        self.assertRaises(ValidationException, record1.validate)

    def test_listof_does_not_allow_none_for_types_typed_list(self):
        record1 = TypedList(list=[5, 4, None])
        self.assertRaises(ValidationException, record1.validate)

    def test_listof_allow_none_for_nullable_marked_typed_list(self):
        record1 = NullableList(list=[5, 4, None])
        record1.validate()

    def test_listof_validate_raises_for_one_item(self):
        quiz = ListQuiz(numbers=[200, 2, 4, 200, 6, 200])
        with self.assertRaises(ValidationException) as context:
            quiz.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['numbers.1'],
                         "value is not greater than or equal 100")

    def test_listof_accepts_dict_value(self):
        val = ListDict(numbers=[{'ab': 1, 'cd': 2}, {'ef': 3}])
        val.validate()
        val = ListDict(numbers=[{'ab': 1, 'cd': 2}, {'ef': True}])
        with self.assertRaises(ValidationException):
            val.validate()



"""module for sqrt modifier."""
from __future__ import annotations
from math import sqrt
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class SqrtModifier(Modifier):
    """Sqrt modifier sqrts number value."""

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is int or type(ctx.val) is float:
            if(ctx.val >= 0):
                return sqrt(ctx.val)
            else:
                ctx.raise_vexc('value is less than 0 thus cannot be sqrted')
        else:
            return ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperRound:
    round_value: Optional[int] = types.round.int
    ceil_value: Optional[int] = types.ceil.int
    floor_value: Optional[int] = types.floor.int



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import UnauthorizedActionException
from tests.classes.gs_article import GSArticle, GSAuthor, GSTArticle
from tests.classes.gm_article import GMArticle, GMAuthor


class TestCanCreate(TestCase):

    def test_guards_raises_if_no_operator_is_assigned(self):
        article = GSArticle(name='P', content='C')
        paid_author = GSAuthor(id='P', name='A', paid_user=True)
        article.author = paid_author
        with self.assertRaises(UnauthorizedActionException):
            article.save()

    def test_guards_are_called_for_new_objects_on_save(self):
        article = GSArticle(name='P', content='C')
        paid_author = GSAuthor(id='P', name='A', paid_user=True)
        article.author = paid_author
        article.opby(paid_author)
        article.save()
        free_author = GSAuthor(id='F', name='A', paid_user=False)
        article.author = free_author
        article.opby(free_author)
        with self.assertRaises(UnauthorizedActionException):
            article.save()

    def test_guards_are_not_called_for_existing_objects_on_save(self):
        article = GSArticle(name='P', content='C')
        setattr(article, '_is_new', False)
        paid_author = GSAuthor(id='P', name='A', paid_user=True)
        article.author = paid_author
        article.opby(paid_author)
        article.save()
        free_author = GSAuthor(id='F', name='A', paid_user=False)
        article.author = free_author
        article.opby(free_author)
        article.save()

    def test_multiple_guards_are_called_for_new_objects_on_save(self):
        article = GMArticle(name='P', content='C')
        paid_author = GMAuthor(id='P', name='A', paid_user=True)
        article.author = paid_author
        article.opby(paid_author)
        article.save()
        free_author = GMAuthor(id='F', name='A', paid_user=False)
        article.author = free_author
        article.opby(free_author)
        with self.assertRaises(UnauthorizedActionException):
            article.save()

    def test_multiple_guards_are_not_called_for_existing_objects_on_save(self):
        article = GMArticle(name='P', content='C')
        setattr(article, '_is_new', False)
        paid_author = GMAuthor(id='P', name='A', paid_user=True)
        article.author = paid_author
        article.opby(paid_author)
        article.save()
        free_author = GMAuthor(id='F', name='A', paid_user=False)
        article.author = free_author
        article.opby(free_author)
        article.save()

    def test_types_guard_is_called_for_new_object_on_save(self):
        article = GSTArticle(name='P', content='C')
        paid_author = GSAuthor(id='P', name='P', paid_user=True)
        article.opby(paid_author)
        article.save()
        free_author = GSAuthor(id='F', name='A', paid_user=False)
        article.author = free_author
        article.opby(free_author)
        with self.assertRaises(UnauthorizedActionException):
            article.save()



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_calculation import SimpleCalculation


class TestDiv(TestCase):


    def test_div_with_int_value_divs_by_original_value_(self):
        div_int = SimpleCalculation(i_div=10)
        self.assertEqual(div_int.i_div, 2)

    def test_div_with_float_value_divs_by_original_value_(self):
        div_float = SimpleCalculation(f_div=7.5)
        self.assertEqual(div_float.f_div, 3)

    def test_div_divs_callable_value_to_original_value(self):
        div_float = SimpleCalculation(c_div=8.4)
        self.assertEqual(div_float.c_div, 2)

    def test_div_divs_types_value_to_original_value(self):
        div_float = SimpleCalculation(t_div=8.4)
        self.assertEqual(div_float.t_div, 2)



from __future__ import annotations
from jsonclasses import jsonclass, types


def create_first_revision(article: CBArticle) -> None:
    CBRevision(name='First').article = article


@jsonclass
class CBRevision:
    name: str
    article: CBArticle = types.objof('CBArticle').linkto.required


@jsonclass(on_create=create_first_revision)
class CBArticle:
    name: str
    content: str
    revisions: list[CBRevision] = types.nonnull.listof('CBRevision') \
                                       .linkedby('article')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.operator import (AsopUser, AsopTeam, AsopdUser, AsopdTeam)


class TestOperator(TestCase):

    def test_asop_assigns_transformed_to_field_if_operator_is_present(self):
        user = AsopUser(name='U', id='id')
        team = AsopTeam(name='T')
        team.opby(user)
        self.assertEqual(team.owner_id, user.id)
        team.validate()

    def test_asop_is_invalid_when_operator_is_not_present_on_create(self):
        team = AsopTeam(name='T')
        with self.assertRaises(ValidationException) as context:
            team.validate()
        self.assertEqual(context.exception.keypath_messages['owner'],
                         "no operator being assigned")

    def test_asop_is_not_invalid_when_updating_other_fields(self):
        team = AsopTeam(name='T')
        setattr(team, '_is_new', False)
        team.validate()

    def test_asop_is_invalid_when_operator_is_not_present_on_update(self):
        team = AsopTeam(name='T')
        setattr(team, '_is_new', False)
        setattr(team, '_modified_fields', ('owner',))
        with self.assertRaises(ValidationException) as context:
            team.validate()
        self.assertEqual(context.exception.keypath_messages['owner'],
                         "no operator being assigned")

    def test_asopd_assigns_directly_to_field_if_operator_is_present(self):
        user = AsopdUser(name='U', id='id')
        team = AsopdTeam(name='T')
        team.opby(user)
        self.assertEqual(team.owner_id, user.id)
        team.validate()

    def test_asopd_is_invalid_when_operator_is_not_present_on_create(self):
        team = AsopdTeam(name='T')
        with self.assertRaises(ValidationException) as context:
            team.validate()
        self.assertEqual(context.exception.keypath_messages['owner'],
                         "no operator being assigned")

    def test_asopd_is_not_invalid_when_updating_other_fields(self):
        team = AsopdTeam(name='T')
        setattr(team, '_is_new', False)
        team.validate()

    def test_asopd_is_invalid_when_operator_is_not_present_on_update(self):
        team = AsopdTeam(name='T')
        setattr(team, '_is_new', False)
        setattr(team, '_modified_fields', ('owner',))
        with self.assertRaises(ValidationException) as context:
            team.validate()
        self.assertEqual(context.exception.keypath_messages['owner'],
                         "no operator being assigned")



"""module for upperbond modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class UpperbondModifier(Modifier):
    """Upperbond modifier returns the value of the field to upperbond if
    the value of the field is larger than upperbond."""

    def __init__(self, max_value: int | float | Callable | Types) -> None:
        self.max_value = max_value

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is int or type(ctx.val) is float:
            max_value = self.resolve_param(self.max_value, ctx)
            return max_value if ctx.val > max_value else ctx.val
        return ctx.val



"""
This module contains `jsonclass`, the decorator for JSON Classes.
"""
from __future__ import annotations
from jsonclasses.keypath import identical_key
from typing import (
    Optional, Union, Callable, TypeVar, overload, cast, TYPE_CHECKING
)
from dataclasses import dataclass
from .jconf import (
    JConf, OnCreate, CanCreate, OnDelete, CanUpdate, CanDelete, CanRead,
    OnUpdate
)
from .jfield import JField
from .cdef import CDef
from .jsonclassify import jsonclassify
from .jobject import JObject
if TYPE_CHECKING:
    from .types import Types


T = TypeVar('T', bound=type)


@overload
def jsonclass(cls: T) -> T | type[JObject]: ...


@overload
def jsonclass(
    cls: None = None,
    class_graph: Optional[str] = 'default',
    output_key_strategy: Optional[Callable[[str], str]] = None,
    input_key_strategy: Optional[Callable[[str], str]] = None,
    camelize_json_keys: Optional[bool] = None,
    strict_input: Optional[bool] = None,
    ref_name_strategy: Optional[Callable[[JField], str]] = None,
    validate_all_fields: Optional[bool] = None,
    abstract: Optional[bool] = None,
    reset_all_fields: Optional[bool] = None,
    output_null: Optional[bool] = None,
    on_create: OnCreate | list[OnCreate] | Types | None = None,
    on_update: OnUpdate | list[OnUpdate] | Types | None = None,
    on_delete: OnDelete | list[OnDelete] | Types | None = None,
    can_create: CanCreate | list[CanCreate] | Types | None = None,
    can_update: CanUpdate | list[CanUpdate] | Types | None = None,
    can_delete: CanDelete | list[CanDelete] | Types | None = None,
    can_read: CanRead | list[CanRead] | Types | None = None,
) -> Callable[[T], T | type[JObject]]: ...


@overload
def jsonclass(
    cls: T,
    class_graph: Optional[str] = 'default',
    output_key_strategy: Optional[Callable[[str], str]] = None,
    input_key_strategy: Optional[Callable[[str], str]] = None,
    camelize_json_keys: Optional[bool] = None,
    strict_input: Optional[bool] = None,
    ref_name_strategy: Optional[Callable[[JField], str]] = None,
    validate_all_fields: Optional[bool] = None,
    abstract: Optional[bool] = None,
    reset_all_fields: Optional[bool] = None,
    output_null: Optional[bool] = None,
    on_create: OnCreate | list[OnCreate] | Types | None = None,
    on_update: OnUpdate | list[OnUpdate] | Types | None = None,
    on_delete: OnDelete | list[OnDelete] | Types | None = None,
    can_create: CanCreate | list[CanCreate] | Types | None = None,
    can_update: CanUpdate | list[CanUpdate] | Types | None = None,
    can_delete: CanDelete | list[CanDelete] | Types | None = None,
    can_read: CanRead | list[CanRead] | Types | None = None,
) -> T | type[JObject]: ...


def jsonclass(
    cls: Optional[T] = None,
    class_graph: Optional[str] = 'default',
    output_key_strategy: Optional[Callable[[str], str]] = None,
    input_key_strategy: Optional[Callable[[str], str]] = None,
    camelize_json_keys: Optional[bool] = None,
    strict_input: Optional[bool] = None,
    ref_name_strategy: Optional[Callable[[JField], str]] = None,
    validate_all_fields: Optional[bool] = None,
    abstract: Optional[bool] = None,
    reset_all_fields: Optional[bool] = None,
    output_null: Optional[bool] = None,
    on_create: OnCreate | list[OnCreate] | Types | None = None,
    on_update: OnUpdate | list[OnUpdate] | Types | None = None,
    on_delete: OnDelete | list[OnDelete] | Types | None = None,
    can_create: CanCreate | list[CanCreate] | Types | None = None,
    can_update: CanUpdate | list[CanUpdate] | Types | None = None,
    can_delete: CanDelete | list[CanDelete] | Types | None = None,
    can_read: CanRead | list[CanRead] | Types | None = None,
) -> Union[Callable[[T], T | type[JObject]], T | type[JObject]]:
    """The jsonclass object class decorator. To declare a jsonclass class, use
    this syntax:

      @jsonclass
      class MyObject:
        my_field_one: str
        my_field_two: bool
    """
    if cls is not None:
        if not isinstance(cls, type):
            raise ValueError('@jsonclass should be used to decorate a class.')
        if camelize_json_keys is False:
            output_key_strategy = identical_key
            input_key_strategy = identical_key
        jconf = JConf(
            cgraph=cast(str, class_graph),
            output_key_strategy=output_key_strategy,
            input_key_strategy=input_key_strategy,
            strict_input=strict_input,
            ref_name_strategy=ref_name_strategy,
            validate_all_fields=validate_all_fields,
            abstract=abstract,
            reset_all_fields=reset_all_fields,
            output_null=output_null,
            on_create=on_create,
            on_update=on_update,
            on_delete=on_delete,
            can_create=can_create,
            can_update=can_update,
            can_delete=can_delete,
            can_read=can_read)
        dcls: type = dataclass(init=False)(cls)
        jcls = jsonclassify(dcls)
        cdef = CDef(jcls, jconf)
        jcls.cdef = cdef
        jconf.cgraph.put(cdef)
        return jcls
    else:
        def parametered_jsonclass(cls):
            return jsonclass(
                cls,
                class_graph=class_graph,
                output_key_strategy=output_key_strategy,
                input_key_strategy=input_key_strategy,
                camelize_json_keys=camelize_json_keys,
                strict_input=strict_input,
                ref_name_strategy=ref_name_strategy,
                validate_all_fields=validate_all_fields,
                abstract=abstract,
                reset_all_fields=reset_all_fields,
                output_null=output_null,
                on_create=on_create,
                on_update=on_update,
                on_delete=on_delete,
                can_create=can_create,
                can_update=can_update,
                can_delete=can_delete,
                can_read=can_read)
        return parametered_jsonclass



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class NullableList:
    name: Optional[str]
    list: list[int] = types.listof(types.nullable.int).required



"""This module contains all modifier markers."""
# flake8: noqa: F401
from .modifier import Modifier
from .invalid_modifier import InvalidModifier

from .primary_modifier import PrimaryModifier

# access markers
from .writeonly_modifier import WriteonlyModifier
from .writeonce_modifier import WriteonceModifier
from .writenonnull_modifier import WriteNonnullModifier
from .readonly_modifier import ReadonlyModifier
from .readwrite_modifier import ReadwriteModifier
from .temp_modifier import TempModifier
from .getter_modifier import GetterModifier
from .setter_modifier import SetterModifier
from .nocopy_modifier import NoCopyModifier

# database index command markers
from .index_modifier import IndexModifier
from .unqueryable_modifier import UnqueryableModifier
from .queryable_modifier import QueryableModifier

# orm relationship command markers

from .embedded_modifier import EmbeddedModifier
from .linkto_modifier import LinkToModifier
from .linkedby_modifier import LinkedByModifier
from .linkedthru_modifier import LinkedThruModifier
from .linkedin_modifier import LinkedInModifier
from .referrer_modifier import ReferrerModifier
from .referee_modifier import RefereeModifier
from .deny_modifier import DenyModifier
from .cascade_modifier import CascadeModifier
from .nullify_modifier import NullifyModifier

# eager validation markers
from .eager_modifier import EagerModifier

# preserialize validation markers
from .preserialize_modifier import PreserializeModifier

# str modifiers
from .str_modifier import StrModifier
from .match_modifier import MatchModifier
from .truncate_modifier import TruncateModifier
from .trim_modifier import TrimModifier
from .url_modifier import UrlModifier
from .email_modifier import EmailModifier
from .hexcolor_modifier import HexColorModifier
from .securepw_modifier import SecurepwModifier
from .digit_modifier import DigitModifier
from .alpha_modifier import AlphaModifier
from .numeric_modifier import NumericModifier
from .alnum_modifier import AlnumModifier
from .totitle_modifier import ToTitleModifier
from .tocap_modifier import ToCapModifier
from .tolower_modifier import ToLowerModifier
from .toupper_modifier import ToUpperModifier
from .replace_modifier import ReplaceModifier
from .replacer_modifier import ReplacerModifier
from .split_modifier import SplitModifier
from .join_modifier import JoinModifier
from .salt_modifier import SaltModifier
from .tostr_modifier import ToStrModifier
from .padstart_modifier import PadStartModifier
from .padend_modifier import PadEndModifier
from .mongoid_modifier import MongoIdModifier

# number modifiers
from .int_modifier import IntModifier
from .float_modifier import FloatModifier
from .min_modifier import MinModifier
from .max_modifier import MaxModifier
from .lt_modifier import LtModifier
from .gt_modifier import GtModifier
from .nonnegative_modifier import NonnegativeModifier
from .nonpositive_modifier import NonpositiveModifier
from .range_modifier import RangeModifier
from .positive_modifier import PositiveModifier
from .negative_modifier import NegativeModifier
from .round_modifier import RoundModifier
from .ceil_modifier import CeilModifier
from .floor_modifier import FloorModifier
from .odd_modifier import OddModifier
from .even_modifier import EvenModifier
from .abs_modifier import AbsModifier
from .tofloat_modifier import ToFloatModifier
from .toint_modifier import ToIntModifier
from .add_modifier import AddModifier
from .sub_modifier import SubModifier
from .mul_modifier import MulModifier
from .div_modifier import DivModifier
from .mod_modifier import ModModifier
from .pow_modifier import PowModifier
from .sqrt_modifier import SqrtModifier
from .map_modifier import MapModifier
from .filter_modifier import FilterModifier
from .upperbond_modifier import UpperbondModifier
from .lowerbond_modifier import LowerbondModifier

# bool modifiers
from .bool_modifier import BoolModifier
from .tobool_modifier import ToBoolModifier
from .inverse_modifier import InverseModifier

# datetime modifiers
from .date_modifier import DateModifier
from .datetime_modifier import DatetimeModifier
from .before_modifier import BeforeModifier
from .after_modifier import AfterModifier
from .tobosec_modifier import ToBoSecModifier
from .tobomin_modifier import ToBoMinModifier
from .tobohour_modifier import ToBoHourModifier
from .tonextsec_modifier import ToNextSecModifier
from .tonextmin_modifier import ToNextMinModifier
from .tonexthour_modifier import ToNextHourModifier
from .tonextyear_modifier import ToNextYearModifier
from .tonextmon_modifier import ToNextMonModifier
from .tonextday_modifier import ToNextDayModifier
from .toboyear_modifier import ToBoYearModifier
from .tobomon_modifier import ToBoMonModifier
from .toboday_modifier import ToBoDayModifier
from .fmtd_modifier import FormatDatetimeModifier

# enum modifiers
from .enum_modifier import EnumModifier
from .inputvalue_modifier import InputValueModifier
from .inputname_modifier import InputNameModifier
from .inputlname_modifier import InputLnameModifier
from .inputall_modifier import InputAllModifier
from .outputvalue_modifier import OutputValueModifier
from .outputname_modifier import OutputNameModifier
from .outputlname_modifier import OutputLnameModifier

# iterable modifiers
from .oneof_modifier import OneOfModifier
from .minlength_modifier import MinlengthModifier
from .maxlength_modifier import MaxlengthModifier
from .length_modifier import LengthModifier
from .reverse_modifier import ReverseModifier
from .hasprefix_modifier import HasPrefixModifier
from .hassuffix_modifier import HasSuffixModifier
from .isprefixof_modifier import IsPrefixOfModifier
from .issuffixof_modifier import IsSuffixOfModifier
from .wrapintolist_modifier import WrapIntoListModifier
from .tolist_modifier import ToListModifier
from .insertat_modifier import InsertAtModifier
from .append_modifier import AppendModifier
from .prepend_modifier import PrependModifier
from .len_modifier import LenModifier

# collection modifiers
from .listof_modifier import ListOfModifier
from .dictof_modifier import DictOfModifier

# object modifiers
from .instanceof_modifier import InstanceOfModifier
from .strict_modifier import StrictModifier

# mixed type modifier
from .union_modifier import UnionModifier
from .any_modifier import AnyModifier

# nullability modifiers
from .required_modifier import RequiredModifier
from .nullable_modifier import NullableModifier
from .present_modifier import PresentModifier
from .presentwith_modifier import PresentWithModifier
from .presentwithout_modifier import PresentWithoutModifier
from .nonnull_modifier import NonnullModifier

# custom modifier
from .validate_modifier import ValidateModifier

# comparing modifier and callback
from .reset_modifier import ResetModifier
from .compare_modifier import CompareModifier

# default transformer
from .default_modifier import DefaultModifier

# transform
from .transform_modifier import TransformModifier

# formatter
from .fmt_modifier import FmtModifier

# operator
from .asop_modifier import AsopModifier
from .asopd_modifier import AsopdModifier
from .canc_modifier import CanCModifier
from .canu_modifier import CanUModifier
from .canr_modifier import CanRModifier

# chained modifier
from .chained_modifier import ChainedModifier

# setonsave setter, onsave callback
from .setonsave_modifier import SetOnSaveModifier
from .fsetonsave_modifier import FSetOnSaveModifier
from .onsave_modifier import OnSaveModifier
from .onupdate_modifier import OnUpdateModifier
from .onwrite_modifier import OnWriteModifier

# jsonclasses internal
from .unresolved_modifier import UnresolvedModifier

# authorization
from .authidentity_modifier import AuthIdentityModifier
from .authby_modifier import AuthByModifier

# calc pipeline
from .passin_modifier import PassinModifier
from .checkpw_modifier import CheckpwModifier
from .random_digits_modifier import RandomDigitsModifier
from .random_alnums_modifier import RandomAlnumsModifier
from .random_alnumpuncs_modifier import RandomAlnumpuncsModifier
from .random_float_modifier import RandomFloatModifier
from .random_int_modifier import RandomIntModifier
from .cross_fetch_modifier import CrossFetchModifier
from .fval_modifier import FValModifier
from .fobj_modifier import FObjModifier
from .eq_modifier import EqModifier
from .neq_modifier import NeqModifier
from .this_modifier import ThisModifier
from .at_modifier import AtModifier
from .assign_modifier import AssignModifier
from .uploader_modifier import UploaderModifier
from .isthis_modifier import IsThisModifier
from .oneisvalid_modifier import OneIsValidModifier
from .isobjof_modifier import IsObjOfModifier
from .isobj_modifier import IsObjModifier
from .getop_modifier import GetOpModifier



"""module for vmsg modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class ValidateModifier(Modifier):
    """On invalid, output message as error message."""

    def __init__(self, validator: Callable | Types, msg: str | None = None) -> None:
        if callable(validator):
            params_len = len(signature(validator).parameters)
            if params_len > 2 or params_len < 1:
                raise ValueError('not a valid modifier')
        self.validator = validator
        self.msg = msg if msg is not None else 'invalid value'
        self.use_msg = msg is not None

    def validate(self, ctx: Ctx) -> None:
        from ..types import Types
        if isinstance(self.validator, Types):
            tresult = self.validator.modifier.transform(ctx)
            try:
                self.validator.modifier.validate(ctx.nval(tresult))
                return
            except ValidationException:
                ctx.raise_vexc(self.msg)
        if ctx.val is None:
            return
        params_len = len(signature(self.validator).parameters)
        if params_len == 1:
            result = self.validator(ctx.val)
        elif params_len == 2:
            result = self.validator(ctx.val, ctx)
        if result is None:
            return
        if result is True:
            return
        if result is False:
            ctx.raise_vexc(self.msg)
        if isinstance(result, str):
            ctx.raise_vexc(self.msg if self.use_msg else result)



"""module for email modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from re import compile, match
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


email_regex = compile(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b')

class EmailModifier(Modifier):
    """Email modifier raises if value is not valid email."""

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and match(email_regex, ctx.val) is None:
            ctx.raise_vexc('value is not email string')



from __future__ import annotations
from unittest import TestCase
from typing import List, Dict, Optional
from jsonclasses import jsonclass, types
from jsonclasses.excs import ValidationException
from tests.classes.simple_book import SimpleBook, SimpleBookN
from tests.classes.linked_profile import LinkedProfileN, LinkedProfile
from tests.classes.simple_language import SimpleLanguage


@jsonclass(class_graph='test_instanceof_22')
class User:
    id: int = types.int.primary
    name: str
    posts: List[Post] = types.listof('Post').linkedby('user').required
    comments: List[Comment] = types.listof('Comment').linkedby('commenter').required


@jsonclass(class_graph='test_instanceof_22')
class Post:
    id: int = types.int.primary
    name: str
    user: User = types.linkto.objof('User').required
    comments: List[Comment] = types.listof('Comment').linkedby('post').required


@jsonclass(class_graph='test_instanceof_22')
class Comment:
    id: int = types.int.primary
    content: str
    post: Post = types.linkto.objof('Post').required
    parent: Optional[Comment] = types.linkto.objof('Comment')
    children: List[Comment] = types.listof('Comment').linkedby('parent').required
    commenter: User = types.linkto.objof('User').required


input = {
    'id': 1,
    'name': 'U1',
    'posts': [
        {
            'id': 1,
            'name': 'P1',
            'comments': [
                {
                    'id': 1,
                    'content': 'C1',
                    'commenter': {
                        'id': 2,
                        'name': 'U2'
                    }
                },
                {
                    'id': 2,
                    'content': 'C2',
                    'parent': {
                        'id': 1,
                        'content': 'C1'
                    },
                    'commenter': {
                        'id': 1,
                        'name': 'U1'
                    }
                },
                {
                    'id': 3,
                    'content': 'C3',
                    'commenter': {
                        'id': 2,
                        'name': 'U2'
                    }
                }
            ]
        },
        {
            'id': 2,
            'name': 'P2',
            'comments': [
                {
                    'id': 4,
                    'content': 'C4',
                    'commenter': {
                        'id': 2,
                        'name': 'U2'
                    }
                },
                {
                    'id': 5,
                    'content': 'C5',
                    'commenter': {
                        'id': 3,
                        'name': 'U3'
                    }
                }
            ]
        }
    ]
}


class TestInstanceOfModifier(TestCase):

    def test_instanceof_modifier_creates_instanceof_designated_class_on_transforming(self):
        @jsonclass(class_graph='test_instanceof_1')
        class Address:
            line1: str = types.str
            line2: str = types.str

        @jsonclass(class_graph='test_instanceof_1')
        class User:
            name: str = types.str
            address: Address = types.objof(Address)
        user = User(**{'name': 'John', 'address': {'line1': 'London', 'line2': 'Road'}})
        self.assertIsInstance(user.address, Address)

    def test_instanceof_modifier_fill_children_defaults(self):
        @jsonclass(class_graph='test_instanceof_1_0')
        class Address:
            line1: str = types.str.default('Line1').required
            line2: str = types.str.default('Line2').required

        @jsonclass(class_graph='test_instanceof_1_0')
        class User:
            name: str = types.str.required
            address: Address = types.objof(Address).required
        user = User(**{'name': 'John', 'address': {}})
        self.assertEqual(user.address.line1, 'Line1')
        self.assertEqual(user.address.line2, 'Line2')

    def test_instanceof_modifier_raises_if_type_doesnt_match(self):
        @jsonclass(class_graph='test_instanceof_1_2')
        class Address:
            line1: str = types.str.required
            line2: str = types.str.required

        @jsonclass(class_graph='test_instanceof_1_2')
        class NewAddress:
            line3: str = types.str.required
            line4: str = types.str.required

        @jsonclass(class_graph='test_instanceof_1_2')
        class User:
            name: str = types.str.required
            address: Address = types.objof(Address).required
        user = User(name='John')
        user.address = NewAddress(line3='1', line4='2')
        self.assertRaisesRegex(
            ValidationException,
            "value is not instance of Address",
            user.validate)

    def test_instanceof_modifier_validates_using_modifier_inside(self):
        @jsonclass(class_graph='test_instanceof_2')
        class Address:
            line1: str = types.str.required
            line2: str = types.str

        @jsonclass(class_graph='test_instanceof_2')
        class User:
            name: str = types.str
            address: Address = types.objof(Address)
        user = User(**{'name': 'John', 'address': {'line2': 'Road'}})
        self.assertRaisesRegex(ValidationException,
            "'address\\.line1': value required",
            user.validate)

    def test_instanceof_modifier_convert_subfields_to_json(self):
        @jsonclass(class_graph='test_instanceof_3')
        class Address:
            line1: str = types.str.required
            line2: str = types.str

        @jsonclass(class_graph='test_instanceof_3')
        class User:
            name: str = types.str
            address: Address = types.objof(Address)
        user = User(**{'name': 'John', 'address': {'line2': 'Road', 'line1': 'OK'}})
        result = user.tojson()
        self.assertEqual(result, {'name': 'John', 'address': {'line1': 'OK', 'line2': 'Road'}})

    def test_instanceof_modifier_creates_instances_inside_list(self):
        @jsonclass(class_graph='test_instanceof_4')
        class Address:
            line1: str = types.str
            line2: str = types.str

        @jsonclass(class_graph='test_instanceof_4')
        class User:
            name: str = types.str
            addresses: List[Address] = types.listof(types.objof(Address))
        user = User(**{'name': 'John', 'addresses': [
            {'line1': 'London', 'line2': 'Road'},
            {'line1': 'Paris', 'line2': 'Road'},
        ]})
        self.assertIsInstance(user.addresses[0], Address)
        self.assertIsInstance(user.addresses[1], Address)
        self.assertEqual(len(user.addresses), 2)
        self.assertEqual(user.addresses[0]._data_dict, {'line1': 'London', 'line2': 'Road'})
        self.assertEqual(user.addresses[1]._data_dict, {'line1': 'Paris', 'line2': 'Road'})

    def test_instanceof_modifier_validates_instances_inside_list(self):
        @jsonclass(class_graph='test_instanceof_5')
        class Address:
            line1: str = types.str.required
            line2: str = types.str.required

        @jsonclass(class_graph='test_instanceof_5')
        class User:
            name: str = types.str
            addresses: List[Address] = types.listof(types.objof(Address))
        user = User(**{'name': 'John', 'addresses': [
            {'line1': 'London'},
            {'line1': 'Paris'},
        ]})
        self.assertRaises(ValidationException, user.validate)

    def test_instanceof_modifier_converts_to_json_inside_list(self):
        @jsonclass(class_graph='test_instanceof_6')
        class Address:
            line1: str = types.str
            line2: str = types.str

        @jsonclass(class_graph='test_instanceof_6')
        class User:
            name: str = types.str
            addresses: List[Address] = types.listof(types.objof(Address))
        user = User(**{'name': 'John', 'addresses': [
            {'line1': 'London', 'line2': 'Road'},
            {'line1': 'Paris', 'line2': 'Road'},
        ]})
        result = user.tojson()
        desired = {'name': 'John', 'addresses': [{'line1': 'London', 'line2': 'Road'}, {'line1': 'Paris', 'line2': 'Road'}]}
        self.assertEqual(result, desired)

    def test_instanceof_modifier_allow_argument_to_be_string(self):
        @jsonclass(class_graph='test_instanceof_7')
        class Post:
            title: str = types.str
            content: str = types.str
            author: User = types.objof('User')

        @jsonclass(class_graph='test_instanceof_7')
        class User:
            name: str = types.str
            posts: List[Post] = types.listof(types.objof('Post'))
        user = User(**{'name': 'John', 'posts': [
            {'title': 'P1', 'content': 'C1'},
            {'title': 'P2', 'content': 'C2'},
        ]})
        self.assertIs(user.posts[0].__class__, Post)
        self.assertIs(user.posts[1].__class__, Post)
        result = user.tojson()
        desired = {
            'name': 'John',
            'posts': [
                {'title': 'P1', 'content': 'C1'},
                {'title': 'P2', 'content': 'C2'}
            ]
        }
        self.assertEqual(result, desired)

    def test_instanceof_works_without_assigning_a_types(self):
        @jsonclass(class_graph='test_instanceof_8')
        class Staff:
            position: str
            user: User

        @jsonclass(class_graph='test_instanceof_8')
        class User:
            name: str = types.str
            staff: Staff
        user = User(**{'name': 'John', 'staff': {'position': 'CEO'}})
        self.assertIsInstance(user.staff, Staff)
        self.assertEqual(user.staff.position, 'CEO')
        staff = Staff(**{'position': 'Developer', 'user': {'name': 'Valy'}})
        self.assertIsInstance(staff.user, User)
        self.assertEqual(staff.user.name, 'Valy')

    def test_instanceof_works_in_list_without_assigning_a_types(self):
        @jsonclass(class_graph='test_instanceof_9')
        class Staff:
            position: str
            users: List[User]

        @jsonclass(class_graph='test_instanceof_9')
        class User:
            name: str
            staffs: List[Staff]
        user = User(**{'name': 'John', 'staffs': [{'position': 'CEO'}, {'position': 'CSO'}]})
        self.assertIsInstance(user.staffs[0], Staff)
        self.assertEqual(user.staffs[0].position, 'CEO')
        staff = Staff(**{'position': 'Developer', 'users': [{'name': 'Valy'}, {'name': 'Jonny'}]})
        self.assertIsInstance(staff.users[0], User)
        self.assertEqual(staff.users[0].name, 'Valy')

    def test_instanceof_works_in_dict_without_assigning_a_types(self):
        @jsonclass(class_graph='test_instanceof_10')
        class Staff:
            position: str
            users: Dict[str, User]

        @jsonclass(class_graph='test_instanceof_10')
        class User:
            name: str
            staffs: Dict[str, Staff]
        user = User(**{'name': 'John', 'staffs': {'a': {'position': 'CEO'}, 'b': {'position': 'CSO'}}})
        self.assertIsInstance(user.staffs['a'], Staff)
        self.assertEqual(user.staffs['a'].position, 'CEO')
        staff = Staff(**{'position': 'Developer', 'users': {'a': {'name': 'Valy'}, 'b': {'name': 'Jonny'}}})
        self.assertIsInstance(staff.users['a'], User)
        self.assertEqual(staff.users['a'].name, 'Valy')

    def test_instance_of_accepts_object(self):

        @jsonclass(class_graph='test_instanceof__1')
        class Staff:
            position: str
            user: User = types.linkto.objof('User').required

        @jsonclass(class_graph='test_instanceof__1')
        class User:
            name: str
            staff: Staff = types.objof('Staff').linkedby('user').required

        staff = Staff(position='CFO')
        user = User(name='John', staff=staff)
        self.assertEqual(user.staff.user, user)

    def test_instanceof_raises_if_strict_specified(self):
        @jsonclass(class_graph='test_instanceof_13')
        class Staff:
            position: str
            user: User = types.objof('User').required

        @jsonclass(class_graph='test_instanceof_13')
        class User:
            name: str
            staff: Staff = types.objof('Staff').strict.required
        with self.assertRaisesRegex(ValidationException, "key is not allowed"):
            User(**{'name': 'John', 'staff': {'position': 'CEO', 'boom': True}})

    def test_instanceof_raises_if_strict_instance(self):
        @jsonclass(class_graph='test_instanceof_14', strict_input=True)
        class Staff:
            position: str
            user: User = types.objof('User').required

        @jsonclass(class_graph='test_instanceof_14')
        class User:
            name: str
            staff: Staff = types.objof('Staff').required
        with self.assertRaisesRegex(ValidationException, "'staff\\.boom': key is not allowed"):
            User(**{'name': 'John', 'staff': {'position': 'CEO', 'boom': True}})

    def test_instanceof_create_circular_ref_for_local_and_foreign_binding(self):

        @jsonclass(class_graph='test_instanceof_15')
        class Staff:
            id: int = types.int.primary
            position: str
            user: User = types.linkto.objof('User').required

        @jsonclass(class_graph='test_instanceof_15')
        class User:
            id: int = types.int.primary
            name: str
            staff: Staff = types.objof('Staff').linkedby('user').required

        user = User(**{'id': 1, 'name': 'John', 'staff': {'id': 1, 'position': 'CEO'}})
        self.assertEqual(user.staff.user, user)

    def test_instanceof_create_circular_ref_for_foreign_and_local_binding(self):

        @jsonclass(class_graph='test_instanceof_16')
        class Staff:
            position: str
            user: User = types.objof('User').linkedby('staff').required

        @jsonclass(class_graph='test_instanceof_16')
        class User:
            name: str
            staff: Staff = types.linkto.objof('Staff').required

        user = User(**{'name': 'John', 'staff': {'position': 'CEO'}})
        self.assertEqual(user.staff.user, user)

    def test_instanceof_create_circular_ref_for_foreign_list_and_local_binding(self):

        @jsonclass(class_graph='test_instanceof_17')
        class Post:
            title: str
            user: User = types.linkto.objof('User').required

        @jsonclass(class_graph='test_instanceof_17')
        class User:
            name: str
            posts: List[Post] = types.listof('Post').linkedby('user').required

        user = User(**{'name': 'John', 'posts': [{'title': 'A'}, {'title': 'B'}]})
        self.assertEqual(user.posts[0].user, user)
        self.assertEqual(user.posts[1].user, user)

    def test_instanceof_create_circular_ref_for_local_list_and_foreign_binding(self):

        @jsonclass(class_graph='test_instanceof_18')
        class Post:
            title: str
            user: User = types.objof('User').linkedby('posts').required

        @jsonclass(class_graph='test_instanceof_18')
        class User:
            name: str
            posts: List[Post] = types.linkto.listof('Post').required

        user = User(**{'name': 'John', 'posts': [{'title': 'A'}, {'title': 'B'}]})
        self.assertEqual(user.posts[0].user, user)
        self.assertEqual(user.posts[1].user, user)

    def test_instanceof_create_circular_ref_for_foreign_item_and_local_list_binding(self):

        @jsonclass(class_graph='test_instanceof_19')
        class Post:
            title: str
            user: User = types.linkto.objof('User').required

        @jsonclass(class_graph='test_instanceof_19')
        class User:
            name: str
            posts: List[Post] = types.listof('Post').linkedby('user').required
        post = Post(**{'title': 'A', 'user': {'name': 'B'}})
        self.assertEqual(post.user.posts[0], post)

    def test_instanceof_create_circular_ref_for_local_list_and_foreign_item_binding(self):

        @jsonclass(class_graph='test_instanceof_20')
        class Post:
            title: str
            user: User = types.objof('User').linkedby('posts').required

        @jsonclass(class_graph='test_instanceof_20')
        class User:
            name: str
            posts: List[Post] = types.linkto.listof('Post').required

        post = Post(**{'title': 'A', 'user': {'name': 'B'}})
        self.assertEqual(post.user.posts[0], post)

    def test_instanceof_create_circular_ref_for_many_to_many(self):

        @jsonclass(class_graph='test_instanceof_21')
        class Book:
            title: str
            users: List[User] = types.listof('User').linkedthru('books').required

        @jsonclass(class_graph='test_instanceof_21')
        class User:
            name: str
            books: List[Book] = types.listof('Book').linkedthru('users').required

        book = Book(**{'title': 'A', 'users': [{'name': 'A'}, {'name': 'B'}]})
        self.assertEqual(book.users[0].books[0], book)
        self.assertEqual(book.users[1].books[0], book)

    def test_instanceof_circular_refs_create_refs_for_same_object(self):
        root_user = User(**input)
        self.assertEqual(root_user.name, 'U1')
        self.assertEqual(root_user.posts[0].name, 'P1')
        self.assertEqual(root_user.posts[1].name, 'P2')
        self.assertEqual(root_user.posts[0].comments[0].content, 'C1')
        self.assertEqual(root_user.posts[0].comments[1].content, 'C2')
        self.assertEqual(root_user.posts[0].comments[2].content, 'C3')
        self.assertEqual(root_user.posts[1].comments[0].content, 'C4')
        self.assertEqual(root_user.posts[1].comments[1].content, 'C5')
        self.assertIs(root_user.posts[0].user, root_user)
        self.assertIs(root_user.posts[1].user, root_user)
        self.assertIs(root_user.posts[0].comments[1].commenter, root_user)
        self.assertIs(root_user.posts[0].comments[1].parent,
                      root_user.posts[0].comments[0])
        self.assertIs(root_user.posts[0].comments[1],
                      root_user.posts[0].comments[0].children[0])
        commenter_u2_0 = root_user.posts[0].comments[0].commenter
        commenter_u2_1 = root_user.posts[0].comments[2].commenter
        commenter_u2_2 = root_user.posts[1].comments[0].commenter
        self.assertIs(commenter_u2_0, commenter_u2_1)
        self.assertIs(commenter_u2_1, commenter_u2_2)
        self.assertEqual(len(commenter_u2_1.comments), 3)
        self.assertIs(commenter_u2_1.comments[0],
                      root_user.posts[0].comments[0])
        self.assertIs(commenter_u2_1.comments[1],
                      root_user.posts[0].comments[2])
        self.assertIs(commenter_u2_1.comments[2],
                      root_user.posts[1].comments[0])
        commenter_u3 = root_user.posts[1].comments[1].commenter
        self.assertIs(root_user.posts[1].comments[1], commenter_u3.comments[0])
        self.assertEqual(1, len(root_user.comments))
        self.assertIs(root_user.comments[0], root_user.posts[0].comments[1])

    def test_instanceof_circular_refs_validate_do_not_infinite_loop(self):
        root_user = User(**input)
        root_user.validate()

    def test_instanceof_circular_refs_tojson_do_not_infinite_loop(self):
        root_user = User(**input)
        json = root_user.tojson(reverse_relationship=False)
        result = {'comments': [{
               'commenterId': 1,
               'content': 'C2',
               'id': 2,
               'parent': {'commenterId': 2,
                          'content': 'C1',
                          'id': 1,
                          'postId': 1},
               'parentId': 1,
               'postId': 1}],
            'id': 1,
            'name': 'U1',
            'posts': [{'comments': [{'children': [{'commenterId': 1,
                                        'content': 'C2',
                                        'id': 2,
                                        'parentId': 1,
                                        'postId': 1}],
                          'commenterId': 2,
                          'content': 'C1',
                          'id': 1,
                          'postId': 1},
                         {
                          'commenterId': 1,
                          'content': 'C2',
                          'id': 2,
                          'parent': {'commenterId': 2,
                                     'content': 'C1',
                                     'id': 1,
                                     'postId': 1},
                          'parentId': 1,
                          'postId': 1},
                         {
                          'commenterId': 2,
                          'content': 'C3',
                          'id': 3,
                          'postId': 1}],
            'id': 1,
            'name': 'P1',
            'userId': 1},
           {'comments': [{
                          'commenterId': 2,
                          'content': 'C4',
                          'id': 4,
                          'postId': 2},
                         {
                          'commenterId': 3,
                          'content': 'C5',
                          'id': 5,
                          'postId': 2}],
            'id': 2,
            'name': 'P2',
            'userId': 1}]}
        self.assertEqual(json, result)

    def test_instanceof_circular_refs_tojson_with_rr_do_not_infinite_loop(self):
        root_user = User(**input)
        json = root_user.tojson(reverse_relationship=True)
        result = {'comments': [{
               'commenter': {'id': 1, 'name': 'U1'},
               'commenterId': 1,
               'content': 'C2',
               'id': 2,
               'parent': {'commenterId': 2,
                          'content': 'C1',
                          'id': 1,
                          'postId': 1},
               'parentId': 1,
               'post': {'comments': [{'commenterId': 2,
                                      'content': 'C1',
                                      'id': 1,
                                      'postId': 1},
                                     {'commenterId': 1,
                                      'content': 'C2',
                                      'id': 2,
                                      'parentId': 1,
                                      'postId': 1},
                                     {'commenterId': 2,
                                      'content': 'C3',
                                      'id': 3,
                                      'postId': 1}],
                        'id': 1,
                        'name': 'P1',
                        'user': {'id': 1, 'name': 'U1'},
                        'userId': 1},
               'postId': 1}],
            'id': 1,
            'name': 'U1',
            'posts': [{'comments': [{'children': [{'commenterId': 1,
                                        'content': 'C2',
                                        'id': 2,
                                        'parentId': 1,
                                        'postId': 1}],
                          'commenter': {'id': 2, 'name': 'U2'},
                          'commenterId': 2,
                          'content': 'C1',
                          'id': 1,
                          'post': {'id': 1, 'name': 'P1', 'userId': 1},
                          'postId': 1},
                         {
                          'commenter': {'id': 1, 'name': 'U1'},
                          'commenterId': 1,
                          'content': 'C2',
                          'id': 2,
                          'parent': {'commenterId': 2,
                                     'content': 'C1',
                                     'id': 1,
                                     'postId': 1},
                          'parentId': 1,
                          'post': {'id': 1, 'name': 'P1', 'userId': 1},
                          'postId': 1},
                         {
                          'commenter': {'id': 2, 'name': 'U2'},
                          'commenterId': 2,
                          'content': 'C3',
                          'id': 3,
                          'post': {'id': 1, 'name': 'P1', 'userId': 1},
                          'postId': 1}],
            'id': 1,
            'name': 'P1',
            'user': {'id': 1, 'name': 'U1'},
            'userId': 1},
           {'comments': [{
                          'commenter': {'id': 2, 'name': 'U2'},
                          'commenterId': 2,
                          'content': 'C4',
                          'id': 4,
                          'post': {'id': 2, 'name': 'P2', 'userId': 1},
                          'postId': 2},
                         {
                          'commenter': {'id': 3, 'name': 'U3'},
                          'commenterId': 3,
                          'content': 'C5',
                          'id': 5,
                          'post': {'id': 2, 'name': 'P2', 'userId': 1},
                          'postId': 2}],
            'id': 2,
            'name': 'P2',
            'user': {'id': 1, 'name': 'U1'},
            'userId': 1}]}
        self.assertEqual(json, result)

    def test_instanceof_returns_empty_for_null_value_by_default(self):
        self.assertEqual(SimpleBook().tojson(), {'published': False})

    def test_instanceof_returns_null_for_null_when_option_is_set(self):
        self.assertEqual(SimpleBookN().tojson(), {'published': False, 'name': None})

    def test_instanceof_returns_empty_for_null_local_key_by_default(self):
        self.assertEqual(LinkedProfile().tojson(), {})

    def test_instanceof_returns_null_for_null_local_key_when_option_is_set(self):
        self.assertEqual(LinkedProfileN().tojson(), {'name': None, 'userId': None, 'user': None})

    def test_partial_returns_empty_for_ignored_fields_even_output_null(self):
        l = SimpleLanguage()
        l._is_partial = True
        l._partial_picks = ['name']
        l.name = 'A'
        self.assertEqual(l.tojson(), {'name': 'A'})



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_book import SimpleBook
from tests.classes.simple_deadline import SimpleDeadline


class TestUpdate(TestCase):

    def test_update_without_arguments_wont_change_anything(self):
        book = SimpleBook(name='Thao Bvê', published=False)
        book.update()
        self.assertEqual(book._data_dict,
                         {'name': 'Thao Bvê', 'published': False})

    def test_update_with_keyed_arguments_updates_value(self):
        book = SimpleBook(name='Thao Bvê', published=False)
        book.update(name='Thao Boê')
        self.assertEqual(book._data_dict,
                         {'name': 'Thao Boê', 'published': False})

    def test_update_set_multiple_values_at_once(self):
        book = SimpleBook(name='Thao Boê', published=False)
        book.update(name='Thao Bɛ', published=True)
        self.assertEqual(book._data_dict,
                         {'name': 'Thao Bɛ', 'published': True})

    def test_update_returns_self_and_is_chained(self):
        book = SimpleBook(name='Thao Boê', published=False)
        book.update(name='C').update(name='P') \
            .update(name='T').update(name='B')
        self.assertEqual(book._data_dict, {'published': False, 'name': 'B'})

    def test_update_does_not_trigger_transform(self):
        deadline = SimpleDeadline()
        deadline.update(ended_at='2020-02-04')
        self.assertEqual(deadline._data_dict,
                         {'ended_at': '2020-02-04', 'message': None})

    def test_update_sets_back_value_to_none(self):
        deadline = SimpleDeadline()
        deadline.update(ended_at='2020-02-04').update(ended_at=None)
        self.assertEqual(
            deadline._data_dict,
            {'ended_at': None, 'message': None})

    def test_update_does_not_auto_convert_camelcase_keys_into_snakecase(self):
        deadline = SimpleDeadline()
        with self.assertRaises(ValueError):
            deadline.update(**{'endedAt': '2020-02-04'})

    def test_update_raises_if_given_key_is_not_allowed(self):
        deadline = SimpleDeadline()
        with self.assertRaises(ValueError) as context:
            deadline.update(**{'name': 'a', 'value': 'b'})
        self.assertRegex(str(context.exception),
                         "'(name|value)', '(value|name)' not allowed in "
                         "SimpleDeadline\\.")



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import JSONClassResetNotEnabledError
from tests.classes.resetable_score import ResetableScore
from tests.classes.simple_article import SimpleArticle


class TestReset(TestCase):

    def test_reset_primitive_field(self):
        score = ResetableScore(name='today', total=95)
        score._mark_not_new()
        score.name = 'yesterday'
        score.reset()
        self.assertEqual(score.name, 'today')
        self.assertEqual(score.is_modified, False)
        self.assertEqual(score.modified_fields, ())

    def test_reset_list(self):
        score = ResetableScore(name='today', total=95, scores=[80, 70, 100])
        score._mark_not_new()
        score.scores = [100, 90, 80]
        score.reset()
        self.assertEqual(score.scores, [80, 70, 100])
        self.assertEqual(score.is_modified, False)
        self.assertEqual(score.modified_fields, ())

    def test_reset_dict(self):
        score = ResetableScore(name='today', total=95, history={'y': 4})
        score._mark_not_new()
        score.history['q'] = 5
        score.reset()
        self.assertEqual(score.history, {'y': 4})
        self.assertEqual(score.is_modified, False)
        self.assertEqual(score.modified_fields, ())

    def test_reset_wont_record_if_no_reset_all_fields_option(self):
        article = SimpleArticle(title='q', content='b')
        article._mark_not_new()
        article.title = 'a'
        article.title = 'b'
        self.assertEqual(article.previous_values, {})

    def test_reset_raises_if_no_reset_all_fields_option(self):
        article = SimpleArticle(title='q', content='b')
        article._mark_not_new()
        article.title = 'a'
        article.title = 'b'
        with self.assertRaises(JSONClassResetNotEnabledError):
            article.reset()



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_range import SimpleRange

class TestRange(TestCase):

    def test_range_does_not_raise_if_value_is_in_range(self):
        i = SimpleRange(i=5)
        i.validate()

    def test_range_does_not_raise_if_value_is_in_callable_range(self):
        i = SimpleRange(ic=5)
        i.validate()

    def test_range_does_not_raise_if_value_is_in_types_range(self):
        i = SimpleRange(it=5)
        i.validate()

    def test_range_raises_if_value_is_not_in_range(self):
        i = SimpleRange(i=11)
        with self.assertRaises(ValidationException) as context:
            i.validate()
        self.assertEqual(context.exception.keypath_messages['i'],
                         "value is not less than or equal 10")



"""module for lowerbond modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class LowerbondModifier(Modifier):
    """Lowerbond modifier returns the value of the field to lowerbond if
    the value of the field is smaller than lowerbond."""

    def __init__(self, min_value: int | float | Callable | Types) -> None:
        self.min_value = min_value

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is int or type(ctx.val) is float:
            min_value = self.resolve_param(self.min_value, ctx)
            return min_value if ctx.val < min_value else ctx.val
        return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_at import SimpleAt


class TestAt(TestCase):
    def test_at_gets_the_index_value_of_original_value(self):
        obj = SimpleAt(a={"name": "ben", "sex": "male"})
        self.assertEqual(obj.a, "ben")

    def test_at_gets_the_index_value_of_original_value_with_callable(self):
        obj = SimpleAt(ca={"name": "ben", "sex": "male"})
        self.assertEqual(obj.ca, "ben")

    def test_at_gets_the_index_value_of_original_value_with_types(self):
        obj = SimpleAt(ta={"name": "ben", "sex": "male"})
        self.assertEqual(obj.ta, "ben")



"""module for union modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any
from ..excs import ValidationException
from .modifier import Modifier
from ..rtypes import rtypes
from ..fdef import FDef, FType
if TYPE_CHECKING:
    from ..ctx import Ctx


class UnionModifier(Modifier):
    """Union type modifier validates value against a list of available types.
    """

    def __init__(self, type_list: list[Any]) -> None:
        self.type_list = type_list

    def define(self, fdef: FDef) -> None:
        fdef._ftype = FType.UNION
        fdef._raw_union_types = [rtypes(t) for t in self.type_list]

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        for types in ctx.fdef.raw_union_types:
            if types.fdef._cdef is None:
                types.fdef._cdef = ctx.owner.__class__.cdef
            try:
                return types.modifier.transform(ctx.alterfdef(types.fdef))
            except:
                continue
        return ctx.val

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        for types in ctx.fdef.raw_union_types:
            if types.fdef._cdef is None:
                types.fdef._cdef = ctx.owner.__class__.cdef
            try:
                types.modifier.validate(ctx.alterfdef(types.fdef))
                return
            except ValidationException:
                continue
        ctx.raise_vexc('value is not of any provided type')

    def serialize(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        for types in ctx.fdef.raw_union_types:
            if types.fdef._cdef is None:
                types.fdef._cdef = ctx.owner.__class__.cdef
            try:
                newctx = ctx.alterfdef(types.fdef)
                types.modifier.validate(newctx)
                return types.modifier.serialize(newctx)
            except ValidationException:
                continue
        return None

    def tojson(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        for types in ctx.fdef.raw_union_types:
            if types.fdef._cdef is None:
                types.fdef._cdef = ctx.owner.__class__.cdef
            try:
                newctx = ctx.alterfdef(types.fdef)
                types.modifier.validate(newctx)
                return types.modifier.tojson(newctx)
            except ValidationException:
                continue
        return None



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleProject:
    name: Optional[str]
    attendees: list[str] = types.listof(types.str.minlength(2)).required



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from classes.super_number import SuperNumber

class TestGt(TestCase):

    def test_gt_doesnt_raise_if_float_value_is_greater_than_arg(self):
        n = SuperNumber(ff_gt=5.6)
        n.validate()

    def test_gt_raises_if_float_value_is_equal_to_arg(self):
        n = SuperNumber(ff_gt=5.5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ffGt'],
                         "value is not greater than 5.5")

    def test_gt_raises_if_float_value_is_less_than_arg(self):
        n = SuperNumber(ff_gt=5.4)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ffGt'],
                         "value is not greater than 5.5")

    def test_gt_doesnt_raise_if_int_value_is_greater_than_arg(self):
        n = SuperNumber(if_gt=6)
        n.validate()

    def test_gt_raises_if_int_value_is_equal_to_arg(self):
        n = SuperNumber(if_gt=5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ifGt'],
                         "value is not greater than 5")

    def test_gt_raises_if_int_value_is_less_than_arg(self):
        n = SuperNumber(if_gt=4)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ifGt'],
                         "value is not greater than 5")

    def test_gt_doesnt_raise_if_float_value_is_greater_than_callable_arg(self):
        n = SuperNumber(fcf_gt=5.6)
        n.validate()

    def test_gt_doesnt_raise_if_float_value_is_greater_than_types_arg(self):
        n = SuperNumber(ftf_gt=5.6)
        n.validate()



from __future__ import annotations
from jsonclasses import jsonclass, types


def check_author(article: GSArticle, operator: GSAuthor) -> bool:
    return article.author.id == operator.id and operator.paid_user


@jsonclass
class GSAuthor:
    id: str
    name: str
    paid_user: bool
    articles: list[GSArticle] = types.listof('GSArticle').linkedby('author') \
                                     .required


@jsonclass(can_create=check_author)
class GSArticle:
    name: str
    content: str
    author: GSAuthor


@jsonclass(can_create=types.getop.fval('name').eq(types.this.fval('name')))
class GSTArticle:
    name: str
    content: str



"""module for oneof modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..types import Types
    from ..ctx import Ctx


class OneOfModifier(Modifier):
    """One of modifier validates value against a list of available values."""

    def __init__(self, choices: list[Any] | Callable | Types) -> None:
        self.choices = choices

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        choices = self.resolve_param(self.choices, ctx)
        if ctx.val is not None and ctx.val not in choices:
            ctx.raise_vexc(f'value is not in choices')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_code import SimpleCode

class TestMinlength(TestCase):

    def test_minlength_does_not_raise_if_length_is_greater_than_or_equal_to_constant_minlength_in_str(self):
        min_code = SimpleCode(min_code='1234')
        min_code.validate()
        min_code = SimpleCode(min_code='123456')
        min_code.validate()

    def test_minlength_does_not_raise_if_length_is_greater_than_or_equal_to_constant_minlength_in_list(self):
        l_min_code = SimpleCode(l_min_code=[1, 2, 3, 4])
        l_min_code.validate()
        l_min_code = SimpleCode(l_min_code=[1, 2, 3, 4, 5])
        l_min_code.validate()

    def test_minlength_raises_if_length_is_less_than_constant_minlength_in_str(self):
        min_code = SimpleCode(min_code='123')
        with self.assertRaises(ValidationException):
            min_code.validate()
        min_code = SimpleCode(min_code='1')
        with self.assertRaises(ValidationException):
            min_code.validate()

    def test_minlength_raises_if_length_is_less_than_constant_minlength_in_list(self):
        l_min_code = SimpleCode(l_min_code=[1, 2, 3])
        with self.assertRaises(ValidationException):
            l_min_code.validate()
        l_min_code = SimpleCode(l_min_code=[1])
        with self.assertRaises(ValidationException):
            l_min_code.validate()

    def test_minlength_does_not_raise_if_length_is_greater_than_or_equal_to_callable_minlength_in_str(self):
        c_min_code = SimpleCode(c_min_code='1234')
        c_min_code.validate()
        c_min_code = SimpleCode(c_min_code='123456')
        c_min_code.validate()

    def test_minlength_does_not_raise_if_length_is_greater_than_or_equal_to_callable_minlength_in_list(self):
        cl_min_code = SimpleCode(cl_min_code=[1, 2, 3, 4])
        cl_min_code.validate()
        cl_min_code = SimpleCode(cl_min_code=[1, 2, 3, 4, 5])
        cl_min_code.validate()

    def test_minlength_raises_if_length_is_less_than_callable_minlength_in_str(self):
        c_min_code = SimpleCode(c_min_code='123')
        with self.assertRaises(ValidationException):
            c_min_code.validate()
        c_min_code = SimpleCode(c_min_code='1')
        with self.assertRaises(ValidationException):
            c_min_code.validate()

    def test_minlength_raises_if_length_is_less_than_callable_minlength_in_list(self):
        cl_min_code = SimpleCode(cl_min_code=[1, 2, 3])
        with self.assertRaises(ValidationException):
            cl_min_code.validate()
        cl_min_code = SimpleCode(cl_min_code=[1])
        with self.assertRaises(ValidationException):
            cl_min_code.validate()

    def test_minlength_does_not_raise_if_length_is_greater_than_or_equal_to_types_minlength_in_str(self):
        t_min_code = SimpleCode(t_min_code='1234')
        t_min_code.validate()
        t_min_code = SimpleCode(t_min_code='123456')
        t_min_code.validate()

    def test_minlength_does_not_raise_if_length_is_greater_than_or_equal_to_types_minlength_in_list(self):
        tl_min_code = SimpleCode(tl_min_code=[1, 2, 3, 4])
        tl_min_code.validate()
        tl_min_code = SimpleCode(tl_min_code=[1, 2, 3, 4, 5])
        tl_min_code.validate()

    def test_minlength_raises_if_length_is_less_than_types_minlength_in_str(self):
        t_min_code = SimpleCode(t_min_code='123')
        with self.assertRaises(ValidationException):
            t_min_code.validate()
        t_min_code = SimpleCode(t_min_code='1')
        with self.assertRaises(ValidationException):
            t_min_code.validate()

    def test_minlength_raises_if_length_is_less_than_types_minlength_in_list(self):
        tl_min_code = SimpleCode(tl_min_code=[1, 2, 3])
        with self.assertRaises(ValidationException):
            tl_min_code.validate()
        tl_min_code = SimpleCode(tl_min_code=[1])
        with self.assertRaises(ValidationException):
            tl_min_code.validate()



"""module for readonly modifier."""
from ..fdef import FDef, WriteRule
from .modifier import Modifier


class ReadonlyModifier(Modifier):
    """Readonly modifier marks a field to be readonly."""

    def define(self, fdef: FDef) -> None:
        fdef._write_rule = WriteRule.NO_WRITE



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_user_account import SimpleUserAccount
from tests.classes.simple_contact import SimpleContact


class TestPresentWithout(TestCase):

    def test_presentwithout_raises_if_both_not_presented(self):
        account = SimpleUserAccount()
        with self.assertRaises(ValidationException) as context:
            account.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['email'],
                         "value is not present without ['phone_number']")

    def test_presentwithout_doesnt_raise_if_one_presented(self):
        SimpleUserAccount(email='a@g.com').validate()
        SimpleUserAccount(phone_number='a@g.com').validate()

    def test_presentwithout_raises_if_non_presented(self):
        contact = SimpleContact()
        with self.assertRaises(ValidationException) as context:
            contact.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['email'],
                         "value is not present without ['phone_no', 'address']")

    def test_presentwithout_doesnt_raise_if_one_of_many_presented(self):
        SimpleContact(email='a@g.com').validate()
        SimpleContact(phone_no='a@g.com').validate()
        SimpleContact(address='a@g.com').validate()



"""module for int modifier."""
from ..fdef import FType
from .type_modifier import TypeModifier


class IntModifier(TypeModifier):
    """Int modifier validates value against int type."""

    def __init__(self):
        super().__init__()
        self.cls = int
        self.ftype = FType.INT
        self.exact_type = True



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class SuperOneOf:
    const: list[str] = types.listof(str).default(['abc', 'def', 'ghi']).required
    valconst: str | None = types.str.oneof(['abc', 'def', 'ghi'])
    valcallable: str | None = types.str.oneof(lambda o: o.const)
    valtypes: str | None = types.str.oneof(types.this.fval('const'))



"""The owner observable list."""
from __future__ import annotations
from typing import Generic, Iterable, Any, Protocol, TypeVar, MutableSequence

T_contra = TypeVar('T_contra', contravariant=True)
_T = TypeVar('_T')


class ListOwner(Protocol[T_contra]):

    def __olist_add__(self,
                      olist: OwnedList,
                      idx: int,
                      val: T_contra) -> None: ...

    def __olist_del__(self, olist: OwnedList, val: T_contra) -> None: ...

    def __olist_sor__(self, olist: OwnedList) -> None: ...

    def __olist_will_change__(self, olist: OwnedList) -> None: ...


def is_list_owner(obj: Any):
    has_add = hasattr(obj, '__olist_add__')
    has_del = hasattr(obj, '__olist_del__')
    has_sor = hasattr(obj, '__olist_sor__')
    has_change = hasattr(obj, '__olist_will_change__')
    return has_add and has_del and has_sor and has_change


class OwnedList(list, MutableSequence[_T], Generic[_T]):

    @property
    def owner(self) -> ListOwner:
        return self._owner

    @owner.setter
    def owner(self, val: ListOwner) -> None:
        self._owner = val

    @property
    def keypath(self) -> str:
        return self._keypath

    @keypath.setter
    def keypath(self, val: str) -> None:
        self._keypath = val

    def append(self, value: _T) -> None:
        self.owner.__olist_will_change__(self)
        super().append(value)
        self.owner.__olist_add__(self, len(self) - 1, value)

    def extend(self, values: Iterable[_T]) -> None:
        curlen = len(self)
        if len(values) > 0:
            self.owner.__olist_will_change__(self)
        super().extend(values)
        for v in values:
            self.owner.__olist_add__(self, curlen, v)
            curlen += 1

    def insert(self, index: int, value: _T) -> None:
        curlen = len(self)
        self.owner.__olist_will_change__(self)
        super().insert(index, value)
        idx = min(curlen, index)
        self.owner.__olist_add__(self, idx, value)

    def remove(self, value: _T) -> None:
        self.owner.__olist_will_change__(self)
        super().remove(value)
        self.owner.__olist_del__(self, value)

    def sort(self, **kwargs) -> None:  # TODO: fix argument type hint
        self.owner.__olist_will_change__(self)
        super().sort(**kwargs)
        self.owner.__olist_sor__(self)

    def clear(self) -> None:
        items = [item for item in self]
        self.owner.__olist_will_change__(self)
        super().clear()
        for item in items:
            self.owner.__olist_del__(self, item)

    def pop(self, *args) -> _T:  # TODO: fix argument type hint
        self.owner.__olist_will_change__(self)
        retval = super().pop(*args)
        self.owner.__olist_del__(self, retval)
        return retval

    def reverse(self) -> None:
        self.owner.__olist_will_change__(self)
        super().reverse()
        self.owner.__olist_sor__(self)

    def __add__(self, x: list[_T]) -> list[_T]:
        return super().__add__(x)

    def __iadd__(self, x: Iterable[_T]) -> OwnedList[_T]:
        if len(x) > 0:
            self.owner.__olist_will_change__(self)
        curlen = len(self)
        retval = super().__iadd__(x)
        for v in x:
            self.owner.__olist_add__(self, curlen, v)
            curlen += 1
        return retval

    def __imul__(self, n: int) -> OwnedList[_T]:
        len_self = 0
        if n <= 0:
            remove_list = list(self)
            add_list: list[_T] = []
        elif n == 1:
            remove_list = []
            add_list = []
        else:
            remove_list = []
            ran = range(0, n - 1)
            add_list = []
            for i in ran:
                for item in self:
                    add_list.append(item)
            len_self = len(self)
        if len(remove_list) > 0 or len(add_list) > 0:
            self.owner.__olist_will_change__(self)
        retval = super().__imul__(n)
        for item in remove_list:
            self.owner.__olist_del__(self, item)
        for item in add_list:
            self.owner.__olist_add__(self, len_self, item)
            len_self += 1
        return retval

    def __setitem__(self, *args) -> None:
        len_self = len(self)
        if isinstance(args[0], int):
            idx: int = args[0]
            if idx < 0:
                idx = len_self + idx
            val = self[idx] if 0 <= idx < len_self else None
            self.owner.__olist_will_change__(self)
            super().__setitem__(*args)
            self.owner.__olist_del__(self, val)
            self.owner.__olist_add__(self, idx, args[1])
        elif isinstance(args[0], slice):
            slc: slice = args[0]
            remove_list = self[slc]
            start = slc.start or 0
            step = slc.step or 1
            if start < 0:
                start = len_self + start
            start = min(max(0, start), len_self)
            self.owner.__olist_will_change__(self)
            super().__setitem__(*args)
            for item in remove_list:
                self.owner.__olist_del__(self, item)
            for item in args[1]:
                self.owner.__olist_add__(self, start, item)
                start += step
        else:
            super().__setitem__(*args)

    def __delitem__(self, *args) -> None:
        if isinstance(args[0], int):
            len_self = len(self)
            idx: int = args[0]
            if idx < 0:
                idx = len_self + idx
            val = self[idx] if 0 <= idx < len_self else None
            self.owner.__olist_will_change__(self)
            super().__delitem__(*args)
            self.owner.__olist_del__(self, val)
        elif isinstance(args[0], slice):
            slc: slice = args[0]
            remove_list = self[slc]
            self.owner.__olist_will_change__(self)
            super().__delitem__(*args)
            for item in remove_list:
                self.owner.__olist_del__(self, item)
        else:
            super().__delitem__(*args)



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_type import SuperType


class TestToList(TestCase):

    def test_tolist_transforms_set_value_into_a_list(self):
        l = SuperType(tl={1, 2, 4, 7, 9})
        self.assertEqual(l.tl, [1, 2, 4, 7, 9])

    def test_tolist_transforms_str_value_into_a_list(self):
        l = SuperType(tl="abc&*(_")
        self.assertEqual(l.tl, ["a", "b", "c", "&", "*", "(", "_"])

    def test_tolist_transforms_tuple_value_into_a_list(self):
        l = SuperType(tl=(1, 2, True, "abc"))
        self.assertEqual(l.tl, [1, 2, True, "abc"])

    def test_tolist_keeps_value_if_type_of_value_is_not_set_str_or_tuple(self):
        l = SuperType(tl=12)
        self.assertEqual(l.tl, 12)




from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase

from tests.classes.super_datetime import SuperDateTime


class TestToboday(TestCase):

    def test_toboday_transforms_datetime_into_the_time_of_beginning_of_day(self):
        d = SuperDateTime(dtbd=datetime(2021, 10, 11, 17, 37, 27,43235))
        self.assertEqual(d.dtbd, datetime(2021,10, 11, 0, 0))

    def test_toboday_transforms_date_into_the_time_of_beginning_of_day(self):
        d = SuperDateTime(dbd=date(2021, 10, 11))
        self.assertEqual(d.dbd, date(2021,10, 11))

    def test_toboday_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(sbd="12345")
        self.assertEqual(s.sbd, "12345")



"""module for fsetonsave modifier."""
from __future__ import annotations
from .setonsave_modifier import SetOnSaveModifier
from ..fdef import FDef


class FSetOnSaveModifier(SetOnSaveModifier):
    """Force set on save modifier updates or sets value on save regardless of
    modified or not.
    """

    def define(self, fdef: FDef) -> None:
        fdef._force_set_on_save = True



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleMath:
    i_pow: Optional[int] = types.int.pow(2)
    f_pow: Optional[float] = types.float.pow(2.5)
    c_pow: Optional[float] = types.float.pow(lambda: 2.5)
    t_pow: Optional[float] = types.float.pow(types.default(2.5))


    i_sqrt: Optional[int] = types.int.sqrt
    f_sqrt: Optional[float] = types.float.sqrt



"""module for or modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types



class OneIsValidModifier(Modifier):
    """One is valid modifier validates with subroutines."""

    def __init__(self, subroutines: list[Callable | Types]) -> None:
        for item in subroutines:
            if callable(item):
                params_len = len(signature(item).parameters)
                if params_len > 2 or params_len < 1:
                    raise ValueError('not a valid or subroutine callable')
        self.subroutines = subroutines

    def validate(self, ctx: Ctx) -> None:
        for item in self.subroutines:
            from ..types import Types
            if isinstance(item, Types):
                tresult = item.modifier.transform(ctx)
                try:
                    item.modifier.validate(ctx.nval(tresult))
                    return
                except ValidationException:
                    continue
            else:
                params_len = len(signature(item).parameters)
                if params_len == 1:
                    result = callable(ctx.val)
                elif params_len == 2:
                    result = callable(ctx.val, ctx)
                if result is None:
                    return
                if result is True:
                    return
        ctx.raise_vexc('none is valid')



"""module for writenonnull modifier."""
from ..fdef import FDef, WriteRule
from .modifier import Modifier


class WriteNonnullModifier(Modifier):
    """Write nonnull modifier marks a field as writenonnull. Fields are
    allowed to be modified. However, nil value won't be set.
    """

    def define(self, fdef: FDef) -> None:
        fdef._write_rule = WriteRule.WRITE_NONNULL



from __future__ import annotations
from typing import Generic, NamedTuple, TypeVar
from unittest import TestCase
from jsonclasses.odict import OwnedDict

KT = TypeVar('KT')
VT = TypeVar('VT')


class AddRecord(NamedTuple):
    odict: OwnedDict
    key: KT
    val: VT


class DelRecord(NamedTuple):
    odict: OwnedDict
    val: VT


class Owner(Generic[KT, VT]):

    def __init__(self):
        self.add_records: list[AddRecord] = []
        self.del_records: list[DelRecord] = []

    def __odict_will_change__(self, olist: OwnedDict) -> None:
        pass

    def __odict_add__(self, odict: OwnedDict, key: KT, val: VT) -> None:
        self.add_records.append(AddRecord(odict, key, val))

    def __odict_del__(self, odict: OwnedDict, val: VT) -> None:
        self.del_records.append(DelRecord(odict, val))


class TestOwnedDict(TestCase):

    def test_owned_dict_get_notified_thru_clear(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.clear()
        self.assertEqual(owned_dict, {})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_dict, 2),
            DelRecord(owned_dict, 3)])

    def test_owned_dict_get_notified_thru_pop(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.pop('a')
        self.assertEqual(owned_dict, {'b': 3})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [DelRecord(owned_dict, 2)])

    def test_owned_dict_get_notified_thru_pop_with_default(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.pop('a', 4)
        self.assertEqual(owned_dict, {'b': 3})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [DelRecord(owned_dict, 2)])

    def test_owned_dict_get_notified_thru_pop_with_default_no_key(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.pop('z', 4)
        self.assertEqual(owned_dict, {'a': 2, 'b': 3})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [])

    def test_owned_dict_pop_undefined_key_raises(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(KeyError, 'z'):
            owned_dict.pop('z')

    def test_owned_dict_pop_zero_arguments_raises(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, ('pop expected at least 1 '
                                                'argument, got 0')):
            owned_dict.pop()

    def test_owned_dict_pop_3_arguments_raises(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, ('pop expected at most 2 '
                                                'arguments, got 3')):
            owned_dict.pop(1, 2, 3)

    def test_owned_dict_get_notified_thru_popitem(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.popitem()
        self.assertEqual(owned_dict, {'a': 2})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [DelRecord(owned_dict, 3)])

    def test_owned_dict_get_notified_thru_setdefault(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.setdefault('c', 20)
        self.assertEqual(owned_dict, {'a': 2, 'b': 3, 'c': 20})
        self.assertEqual(owner.add_records, [AddRecord(owned_dict, 'c', 20)])
        self.assertEqual(owner.del_records, [])

    def test_owned_dict_get_notified_thru_setdefault_no_set(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 2, 'b': 3})
        owned_dict.owner = owner
        owned_dict.setdefault('a', 5)
        self.assertEqual(owned_dict, {'a': 2, 'b': 3})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [])

    def test_owned_dict_setdefault_3_arguments_raises(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, ('setdefault expected at most 2'
                                                ' arguments, got 3')):
            owned_dict.setdefault(1, 2, 3)

    def test_owned_dict_setdefault_0_arguments_raises(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, ('setdefault expected at least'
                                                ' 1 argument, got 0')):
            owned_dict.setdefault()

    def test_owned_dict_get_notified_thru_update_map(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        owned_dict.update({'c': 0, 'd': 4, 'e': 5})
        self.assertEqual(owned_dict, {'a': 1, 'b': 2, 'c': 0, 'd': 4, 'e': 5})
        self.assertEqual(owner.add_records, [
            AddRecord(owned_dict, 'c', 0),
            AddRecord(owned_dict, 'd', 4),
            AddRecord(owned_dict, 'e', 5)])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_dict, 3)])

    def test_owned_dict_get_notified_thru_update_iterable(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        owned_dict.update([('c', 0), ('d', 4), ('e', 5)])
        self.assertEqual(owned_dict, {'a': 1, 'b': 2, 'c': 0, 'd': 4, 'e': 5})
        self.assertEqual(owner.add_records, [
            AddRecord(owned_dict, 'c', 0),
            AddRecord(owned_dict, 'd', 4),
            AddRecord(owned_dict, 'e', 5)])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_dict, 3)])

    def test_owned_dict_get_notified_thru_update_map_and_kwargs(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        owned_dict.update({'c': 0, 'd': 4, 'e': 5}, c=7)
        self.assertEqual(owned_dict, {'a': 1, 'b': 2, 'c': 7, 'd': 4, 'e': 5})
        self.assertEqual(owner.add_records, [
            AddRecord(owned_dict, 'c', 7),
            AddRecord(owned_dict, 'd', 4),
            AddRecord(owned_dict, 'e', 5)])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_dict, 3)])

    def test_owned_dict_get_notified_thru_update_iterable_and_kwargs(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        owned_dict.update([('c', 0), ('d', 4), ('e', 5)], c=50)
        self.assertEqual(owned_dict, {'a': 1, 'b': 2, 'c': 50, 'd': 4, 'e': 5})
        self.assertEqual(owner.add_records, [
            AddRecord(owned_dict, 'c', 50),
            AddRecord(owned_dict, 'd', 4),
            AddRecord(owned_dict, 'e', 5)])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_dict, 3)])

    def test_owned_dict_raises_when_wrong_arguments_to_update(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, ('update expected at most 1 '
                                                'argument, got 2')):
            owned_dict.update([('c', 0), ('d', 4), ('e', 5)], {}, c=50)

    def test_owned_dict_raises_when_wrong_type_arguments_to_update(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, ("'int' object is not "
                                                "iterable")):
            owned_dict.update(5)

    def test_owned_dict_get_notified_thru_subscript_set_override(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2, 'c': 3})
        owned_dict.owner = owner
        owned_dict['a'] = 5
        self.assertEqual(owned_dict, {'a': 5, 'b': 2, 'c': 3})
        self.assertEqual(owner.add_records, [AddRecord(owned_dict, 'a', 5)])
        self.assertEqual(owner.del_records, [DelRecord(owned_dict, 1)])

    def test_owned_dict_get_notified_thru_subscript_set_new(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2})
        owned_dict.owner = owner
        owned_dict['c'] = 3
        self.assertEqual(owned_dict, {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(owner.add_records, [AddRecord(owned_dict, 'c', 3)])
        self.assertEqual(owner.del_records, [])

    def test_owned_dict_get_notified_thru_subscript_del(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2})
        owned_dict.owner = owner
        del owned_dict['b']
        self.assertEqual(owned_dict, {'a': 1})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [DelRecord(owned_dict, 2)])

    def test_owned_dict_raises_if_subscript_del_not_exist(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2})
        owned_dict.owner = owner
        with self.assertRaisesRegex(KeyError, 'q'):
            del owned_dict['q']
        self.assertEqual(owned_dict, {'a': 1, 'b': 2})
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [])

    def test_owned_dict_get_notified_thru_or_equal_sign(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2})
        owned_dict.owner = owner
        owned_dict |= {'b': 3, 'c': 4}
        self.assertEqual(owned_dict, {'a': 1, 'b': 3, 'c': 4})
        self.assertEqual(owner.add_records, [
            AddRecord(owned_dict, 'b', 3),
            AddRecord(owned_dict, 'c', 4)])
        self.assertEqual(owner.del_records, [DelRecord(owned_dict, 2)])

    def test_owned_dict_or_equal_sign_raises_if_wrong_argument(self):
        owner = Owner()
        owned_dict = OwnedDict({'a': 1, 'b': 2})
        owned_dict.owner = owner
        with self.assertRaisesRegex(TypeError, "'int' object is not iterable"):
            owned_dict |= 5



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types

@jsonclass
class ZeroNumber:
    inegative: Optional[int] = types.int.negative
    ipositive: Optional[int] = types.int.positive
    inonnegative: Optional[int] = types.int.nonnegative
    inonpositive: Optional[int] = types.int.nonpositive
    fnegative: Optional[float] = types.float.negative
    fpositive: Optional[float] = types.float.positive
    fnonnegative: Optional[float] = types.float.nonnegative
    fnonpositive: Optional[float] = types.float.nonpositive



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperStr:
    password: Optional[str] = types.str.salt
    pads: Optional[str] = types.str.padstart("s", 10)
    padis: Optional[int] = types.int.padstart("s", 10)
    pade: Optional[str] = types.str.padend("e", 10)
    padie: Optional[int] = types.int.padend("e", 10)

    padce: Optional[str] = types.str.padend(lambda: "e", lambda: 10)
    padte: Optional[str] = types.str.padend(types.default("e"), types.default(10))

    c_pads: Optional[str] = types.str.padstart(lambda: "s", lambda: 10)
    t_pads: Optional[str] = types.str.padstart(types.default("s"), types.default(10))

    securepw: Optional[str] = types.str.securepw



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass


@jsonclass(reset_all_fields=True)
class ResetableScore:
    name: str
    total: Optional[int]
    scores: list[float]
    history: dict[str, float]



from __future__ import annotations
from typing import Optional, Union
from datetime import datetime
from jsonclasses import jsonclass, types


@jsonclass
class SimpleMixed:
    name: Optional[str]
    mixed: Optional[Union[int, str]] = types.union([str, int])


@jsonclass
class SimpleMixedT:
    value: int | str = types.union([
        types.str.append('123'),
        types.int.add(123)
    ])


@jsonclass
class SimpleMixedU:
    value: datetime | str
